#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2021-04-29
# @Filename: tools.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

import os
import tempfile

import numpy
import peewee
from astropy.table import Table

from sdssdb.utils.ingest import create_model_from_table

from target_selection.cartons import BaseCarton
from target_selection.exceptions import TargetSelectionError
from target_selection.utils import vacuum_table


def get_file_carton(filename):
    """Returns a carton class that creates a carton based on a FITS file.

    The FITS file is located in the ``open_fiber_path`` which is specified in
    ``python/config/target_selection.yml``.

    The list of FITS files to be loaded is specified in the
    file ``open_fiber_file_list.txt`` which is in the directory ``open_fiber_path``.

    """

    # Import this here to prevent this module not being importable if the database
    # connection is not ready.
    from sdssdb.peewee.sdss5db.catalogdb import (
        Catalog,
        CatalogToGaia_DR2,
        CatalogToGaia_DR3,
        CatalogToLegacy_Survey_DR8,
        CatalogToLegacy_Survey_DR10,
        CatalogToPanstarrs1,
        CatalogToTwoMassPSC,
        Gaia_DR2,
        Gaia_DR3,
        Legacy_Survey_DR8,
        Legacy_Survey_DR10,
        Panstarrs1,
        TwoMassPSC,
    )

    class FileCarton(BaseCarton):
        can_offset = None  # Will be set in query.

        def __init__(self, targeting_plan, config_file=None, schema=None, table_name=None):
            self._file_path = filename

            self._table = Table.read(self._file_path)
            if self._table.masked:
                self._table = self._table.filled()

            self._run_sanity_checks()

            super().__init__(
                targeting_plan,
                config_file=config_file,
                schema=schema,
                table_name=table_name,
            )

            self._disable_query_log = True

        def _run_sanity_checks(self):
            """Runs a series of sanity checks on the FITS table."""

            unique_cartonname = numpy.unique(self._table["cartonname"])
            if len(unique_cartonname) == 1:
                self.name = unique_cartonname[0].lower()
            else:
                raise TargetSelectionError(
                    "Error in get_file_carton(): "
                    + filename
                    + " contains more than one cartonname"
                )

            unique_can_offset = numpy.unique(self._table["can_offset"])
            if len(unique_can_offset) > 1:
                raise TargetSelectionError(
                    "Error in get_file_carton(): "
                    + filename
                    + " contains more than one"
                    + " value of can_offset:"
                    + " can_offset values must be "
                    + " all 0 or all 1"
                )

            if (unique_can_offset[0] != 1) and (unique_can_offset[0] != 0):
                raise TargetSelectionError(
                    "Error in get_file_carton(): "
                    + filename
                    + " can_offset can only be 0 or 1."
                    + " can_offset is "
                    + str(unique_can_offset[0])
                )

            unique_inertial = numpy.unique(self._table["inertial"])
            if len(unique_inertial) > 2:
                raise TargetSelectionError(
                    "Error in get_file_carton(): "
                    + filename
                    + " contains more than two"
                    + " values of inertial:"
                    + " inertial values must be "
                    + " 0 or 1"
                )

            if (unique_inertial[0] != 1) and (unique_inertial[0] != 0):
                raise TargetSelectionError(
                    "Error in get_file_carton(): "
                    + filename
                    + " inertial can only be 0 or 1."
                    + " inertial is "
                    + str(unique_inertial[0])
                )

            # If there is only one inertial value then the above statement
            # is enough. Otherwise, we need to run the below check.
            if len(unique_inertial) == 2:
                if (unique_inertial[1] != 1) and (unique_inertial[1] != 0):
                    raise TargetSelectionError(
                        "Error in get_file_carton(): "
                        + filename
                        + " inertial can only be 0 or 1."
                        + " inertial is "
                        + str(unique_inertial[1])
                    )

            # The valid_program list is from the output of the below command.
            # select distinct(program) from targetdb.carton order by program;
            #
            # mwm_bin is for future mwm binary star cartons
            valid_program = [
                "bhm_aqmes",
                "bhm_csc",
                "bhm_filler",
                "bhm_rm",
                "bhm_spiders",
                "commissioning",
                "mwm_bin",
                "mwm_cb",
                "mwm_dust",
                "mwm_erosita",
                "mwm_filler",
                "mwm_galactic",
                "mwm_gg",
                "mwm_halo",
                "mwm_legacy",
                "mwm_magcloud",
                "mwm_ob",
                "mwm_planet",
                "mwm_rv",
                "mwm_snc",
                "mwm_tessob",
                "mwm_tessrgb",
                "mwm_validation",
                "mwm_wd",
                "mwm_yso",
                "open_fiber",
                "ops",
                "ops_sky",
                "ops_std",
                "SKY",
            ]

            # The valid_category list is from CartonImportTable.pdf
            valid_category = [
                "science",
                "standard_apogee",
                "standard_boss",
                "guide",
                "sky_boss",
                "sky_apogee",
                "standard",
                "sky",
                "veto location boss",
                "veto_location_apogee",
            ]

            # The valid_category list is from CartonImportTable.pdf
            valid_mapper = ["", "MWM", "BHM"]

            unique_category = numpy.unique(self._table["category"])
            if len(unique_category) == 1:
                self.category = unique_category[0].lower()
                if self.category not in valid_category:
                    raise TargetSelectionError(
                        "Error in get_file_carton(): "
                        + filename
                        + " contains invalid category = "
                        + self.category
                    )
            else:
                raise TargetSelectionError(
                    "Error in get_file_carton(): " + filename + " contains more than one category"
                )

            unique_program = numpy.unique(self._table["program"])
            if len(unique_program) == 1:
                self.program = unique_program[0].lower()
                if self.program not in valid_program:
                    raise TargetSelectionError(
                        "Error in get_file_carton(): "
                        + filename
                        + " contains invalid program = "
                        + self.program
                    )
            else:
                raise TargetSelectionError(
                    "Error in get_file_carton(): " + filename + " contains more than one program"
                )

            unique_mapper = numpy.unique(self._table["mapper"])
            if len(unique_mapper) == 1:
                # We do not use lower() for mapper because
                # allowed values for mapper are '' or 'MWM' or 'BHM'.
                self.mapper = unique_mapper[0]
                if self.mapper not in valid_mapper:
                    raise TargetSelectionError(
                        "Error in get_file_carton(): "
                        + filename
                        + " contains invalid mapper = "
                        + self.mapper
                    )
                if self.mapper == "":
                    self.mapper = None
            else:
                raise TargetSelectionError(
                    "Error in get_file_carton(): " + filename + " contains more than one mapper"
                )

            basename_fits = os.path.basename(filename)
            basename_parts = os.path.splitext(basename_fits)
            basename = basename_parts[0]
            carton_name_from_filename = basename.lower()

            if self.name != carton_name_from_filename:
                raise TargetSelectionError(
                    "filename parameter of get_file_carton() and "
                    + "cartonname in FITS file do not match."
                    + "\n"
                    + "carton_name_from_filename = "
                    + carton_name_from_filename
                    + " cartonname = "
                    + self.name
                )

        def copy_data(self, temp_table: str):
            """Copy the input file data to a temporary table.

            The schema of the file carton table is such that we can dump to
            a CSV file without issues.

            """

            temp_csv = tempfile.NamedTemporaryFile(suffix=".csv", delete=False)
            self._table.write(temp_csv.name, format="csv", overwrite=True)

            cursor = self.database.cursor()
            temp_csv.seek(0)
            cursor.copy_expert(f"COPY {temp_table} FROM STDOUT WITH CSV HEADER", temp_csv)
            self.database.commit()

        def build_query(self, version_id, query_region=None):
            self.log.debug(f"Processing file {self._file_path}.")

            # We need to copy the data to a temporary table so that we can
            # join on it. We could use a Peewee ValueList but for large tables
            # that will hit the limit of 1GB in PSQL.

            # Create model for temporary table from FITS table columns.
            # This works fine because we know there are no arrays.
            temp_table = f"inputs_{self.name.lower()}_temp"
            self.database.execute_sql(f"DROP TABLE IF EXISTS {temp_table};")
            temp = create_model_from_table(temp_table, self._table)
            temp._meta.database = self.database
            temp.create_table(temporary=True)

            # Copy the data
            self.copy_data(temp_table)

            self.database.execute_sql(f'CREATE INDEX ON "{temp_table}" ("Gaia_DR3_Source_ID")')
            self.database.execute_sql(f'CREATE INDEX ON "{temp_table}" ("Gaia_DR2_Source_ID")')
            self.database.execute_sql(f'CREATE INDEX ON "{temp_table}" ("LegacySurvey_DR10_ID")')
            self.database.execute_sql(f'CREATE INDEX ON "{temp_table}" ("LegacySurvey_DR8_ID")')
            self.database.execute_sql(f'CREATE INDEX ON "{temp_table}" ("PanSTARRS_DR2_ID")')
            self.database.execute_sql(f'CREATE INDEX ON "{temp_table}" ("TwoMASS_ID")')
            vacuum_table(self.database, temp_table, vacuum=False, analyze=True)

            inertial_case = peewee.Case(
                None,
                ((temp.inertial.cast("boolean").is_null(), False),),
                temp.inertial.cast("boolean"),
            )

            # List of columns and aliases for the final query table.
            query_columns = [
                Catalog.catalogid,
                temp.Gaia_DR3_Source_ID.alias("gaia_dr3_source_id"),
                temp.Gaia_DR2_Source_ID.alias("gaia_source_id"),
                temp.LegacySurvey_DR10_ID.alias("ls_id10"),
                temp.LegacySurvey_DR8_ID.alias("ls_id8"),
                temp.PanSTARRS_DR2_ID.alias("catid_objid"),
                temp.TwoMASS_ID.alias("designation"),
                Catalog.ra,
                Catalog.dec,
                temp.delta_ra.cast("double precision"),
                temp.delta_dec.cast("double precision"),
                inertial_case.alias("inertial"),
                temp.cadence,
                temp.priority,
                temp.instrument,
                temp.can_offset.cast("boolean").alias("can_offset"),
                peewee.Value(0).alias("value"),
            ]

            # Calculate number of rows in the table for each parent catalogue identifier and
            # run some sanity checks.
            len_table = len(self._table)
            len_gaia_dr3 = len(self._table[self._table["Gaia_DR3_Source_ID"] > 0])
            len_gaia_dr2 = len(self._table[self._table["Gaia_DR2_Source_ID"] > 0])
            len_legacysurvey_dr10 = len(self._table[self._table["LegacySurvey_DR10_ID"] > 0])
            len_legacysurvey_dr8 = len(self._table[self._table["LegacySurvey_DR8_ID"] > 0])
            len_panstarrs_dr2 = len(self._table[self._table["PanSTARRS_DR2_ID"] > 0])

            # TwoMass_ID corresponds to the designation column of
            # the table catalogdb.twomass_psc.
            # Since the designation column is a text column, below
            # we are comparing it to the string 'NA' and not the integer 0.
            len_twomass_psc = len(self._table[self._table["TwoMASS_ID"] != "NA"])

            # Make sure this is not an empty table.
            if len_table == 0:
                raise TargetSelectionError(
                    f"Error in get_file_carton(): {self._file_path} is an empty table"
                )

            # There must be exactly one non-zero id per row else raise an exception.
            if (
                len_gaia_dr3
                + len_gaia_dr2
                + len_legacysurvey_dr10
                + len_legacysurvey_dr8
                + len_panstarrs_dr2
                + len_twomass_psc
            ) != len_table:
                raise TargetSelectionError(
                    "Error in get_file_carton(): "
                    + "(len_gaia_dr3 + len_gaia_dr2 + "
                    + "len_legacysurvey_dr10 + len_legacysurvey_dr8 +"
                    + "len_panstarrs_dr2 + len_twomass_psc) != "
                    + "len_table"
                )

            # For each identifier that has non-zero targets, appends some information that
            # we need to create the subquery for that table. This includes the CatalogToX model
            # to which we need to join, the field on which to join in the parent catalogue,
            # and the column in the temporary table. In all cases except 2MASS, we join on the
            # parent catalogue primary key.
            model_data = []

            if len_gaia_dr3 > 0:
                model_data.append(
                    {
                        "catalog_to": CatalogToGaia_DR3,
                        "parent_field": Gaia_DR3.source_id,
                        "temp_column": "Gaia_DR3_Source_ID",
                    }
                )

            if len_gaia_dr2 > 0:
                model_data.append(
                    {
                        "catalog_to": CatalogToGaia_DR2,
                        "parent_field": Gaia_DR2.source_id,
                        "temp_column": "Gaia_DR2_Source_ID",
                    }
                )

            if len_legacysurvey_dr10 > 0:
                model_data.append(
                    {
                        "catalog_to": CatalogToLegacy_Survey_DR10,
                        "parent_field": Legacy_Survey_DR10.ls_id,
                        "temp_column": "LegacySurvey_DR10_ID",
                    }
                )

            if len_legacysurvey_dr8 > 0:
                model_data.append(
                    {
                        "catalog_to": CatalogToLegacy_Survey_DR8,
                        "parent_field": Legacy_Survey_DR8.ls_id,
                        "temp_column": "LegacySurvey_DR8_ID",
                    }
                )

            if len_panstarrs_dr2 > 0:
                model_data.append(
                    {
                        "catalog_to": CatalogToPanstarrs1,
                        "parent_field": Panstarrs1.catid_objid,
                        "temp_column": "PanSTARRS_DR2_ID",
                    }
                )

            if len_twomass_psc > 0:
                model_data.append(
                    {
                        "catalog_to": CatalogToTwoMassPSC,
                        "parent_field": TwoMassPSC.designation,
                        "temp_column": "TwoMASS_ID",
                    }
                )

            if len(model_data) == 0:
                raise TargetSelectionError(
                    "Error in get_file_carton(): no join model found for the file carton."
                )

            # Create a query for each join model. The final query will be the union of all.
            # For each join model, we need to account for cases in which an identifier is
            # associated with more than one catalogid via a window function (we select either
            # the phase 1 match, or the one with the smallest distance).

            queries = []

            for data in model_data:
                catalog_to_model = data["catalog_to"]
                parent_field = data["parent_field"]
                temp_column = data["temp_column"]

                # Get the model class field for the column in the temporary table with the
                # identifier for this case.
                temp_field = getattr(temp, temp_column)

                # Get the psrent model. The only reason why we need to join all the way to the
                # parent catalogue model is 2MASS for which the column TwoMASS_ID in the file
                # carton corresponds to the designation column in the TwoMassPSC model, which
                # is not the primary key.
                parent_model = parent_field.model

                # Create a subquery that ranks the rows by distance to the target. Since
                # temp is also used in the main query, we need to alias it.
                # Note that we are using ROW_NUMBER() and not RANK() because the latter would
                # assign the same rank to multiple rows with the same distance. This can happen
                # if two catalogids are associated with a target in phase 1 (an example of this
                # is a Gaia DR2 target that has been deblended into two Gaia DR3 targets).
                # We also order by catalogid to ensure that the query is deterministic but
                # ultimately we are randomly selecting one of the targets here.
                temp_alias = temp.alias("temp_alias")
                temp_alias_field = getattr(temp_alias, temp_column)

                distance_rank_partition = peewee.fn.row_number().over(
                    partition_by=[catalog_to_model.target_id],
                    order_by=[
                        peewee.fn.coalesce(catalog_to_model.distance, 0.0).asc(),
                        catalog_to_model.catalogid.asc(),
                    ],
                )

                sub_query = (
                    temp_alias.select(
                        catalog_to_model.catalogid,
                        parent_field.alias("target_id"),
                        distance_rank_partition.alias("distance_rank"),
                    )
                    .join(parent_model, on=(temp_alias_field == parent_field))
                    .join(catalog_to_model)
                    .where(
                        catalog_to_model.best >> True,
                        catalog_to_model.version_id == version_id,
                    )
                ).alias("distance_rank_subquery")

                # Now add the main query. We join the subquery to the temporary table to
                # get all the relevant columns, but keep only the entries with distance_rank=1.
                queries.append(
                    Catalog.select(*query_columns, sub_query.c.distance_rank)
                    .join(
                        sub_query,
                        on=(Catalog.catalogid == sub_query.c.catalogid),
                    )
                    .join(
                        temp,
                        on=(temp_field == sub_query.c.target_id),
                    )
                    .where(sub_query.c.distance_rank == 1)
                    .distinct(Catalog.catalogid)
                )

            # Union all queries.
            query_union = queries[0]
            for query in queries[1:]:
                query_union = query_union | query

            # It seems to work better if we disable sequential scans and force the use of the
            # indices.
            self.database.execute_sql("SET LOCAL enable_seqscan = off;")

            # Now just distinct on catalogid for all the unions. Although we already have a
            # distinct in each query, they can yield the same catalogid from different queries.
            query_union = query_union.cte("query_union")

            return (
                query_union.select(query_union.__star__)
                .distinct(query_union.c.catalogid)
                .with_cte(query_union)
            )

        def post_process(self, model, **kwargs):
            """Runs sanity checks on the output of the query."""

            n_file_table = len(self._table)
            n_query = model.select().count()

            if n_file_table != n_query:
                self.log.warning(
                    f"The number of rows in the file table ({n_file_table}) does not "
                    f"match the number of rows returned by the query ({n_query})."
                )

    return FileCarton


def create_table_as(
    query,
    table_name,
    schema=None,
    temporary=False,
    database=None,
    execute=True,
    overwrite=False,
    indices=[],
    analyze=True,
):
    """Creates a table from a query.

    Parameters
    ----------
    query
        A Peewee ``ModelSelect`` or a string with the query to create a table from.
    table_name
        The name of the table to create.
    schema
        The schema in which to create the table. If ``table_name`` is in the
        form ``schema.table``, the schema parameter is overridden by ``table_name``.
    temporary
        Whether to create a temporary table instead of a persistent one.
    database
        The database connection to use to execute the query. If not passed and the
        query is a ``ModelSelect``, the database will be inherited from the query
        model.
    execute
        Whether to actually execute the query. Requires ``database`` to be passed.
    overwrite
        If `True`, the table will be create even if a table with the same name
        already exists. Requires ``database`` or will be ignored.
    analyze
        Whether to ``VACUUM ANALIZE`` the new table.
        Only relevant if ``execute=True``.
    indices
        List of columns to create indices on. Only relevant if ``execute=True``.

    Returns
    -------
    create_query
        A tuple in which the first element is a Peewee ``Table`` for the created table
        (the table is bound to ``database`` if passed), and the ``CREATE TABLE AS``
        query as a string.

    """

    if "." in table_name:
        schema, table_name = table_name.split(".")

    if schema is None and temporary is False:
        schema = "public"
    elif temporary is True:
        schema = None

    path = f"{schema}.{table_name}" if schema else table_name
    create_sql = f'CREATE {"TEMPORARY " if temporary else ""}TABLE {path} AS '

    if database is None and isinstance(query, peewee.ModelSelect):
        database = query.model._meta.database

    if overwrite and database:
        database.execute_sql(f"DROP TABLE IF EXISTS {path};")

    query_sql, params = database.get_sql_context().sql(query).query()
    cursor = database.cursor()
    query_str = cursor.mogrify(query_sql, params).decode()

    if execute:
        if database is None:
            raise RuntimeError("Cannot execute query without a database.")

        with database.atomic():
            database.execute_sql(create_sql + query_sql, params)

        for index in indices:
            if isinstance(index, (list, tuple)):
                index = ",".join(index)
            database.execute_sql(f"CREATE INDEX ON {path} ({index})")

        if analyze:
            database.execute_sql(f"VACUUM ANALYZE {path}")

    table = peewee.Table(table_name, schema=schema).bind(database)

    create_str = create_sql + query_str
    return table, create_str



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: Pramod Gupta (psgupta@uw.edu)
# @Date: 2022-12-11
# @Filename: mwm_wd.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

import peewee
from peewee import fn

from sdssdb.peewee.sdss5db.catalogdb import (
    Catalog,
    CatalogToGaia_DR3,
    Gaia_DR3,
    Gedr3spur_main,
    WD_gaia_dr3,
)

from target_selection.cartons import BaseCarton


class MWM_WD_PWD_boss_Carton(BaseCarton):
    """MWM White Dwarfs.

    Definition:

    Reference: Gentile Fusillo et al. 2021
    A catalogue of white dwarfs in Gaia EDR3.

    select all targets from table wd_gaia_d3
    where Pwd > 0.5 and Gmag <= 20.

    (above Gmag means gmag_vega)

    """

    name = "mwm_wd_pwd_boss"
    mapper = "MWM"
    category = "science"
    program = "mwm_wd"
    instrument = "BOSS"
    cadence = "dark_2x2"
    priority = 1400
    can_offset = True

    def build_query(self, version_id, query_region=None):
        query = (
            CatalogToGaia_DR3.select(
                CatalogToGaia_DR3.catalogid,
                WD_gaia_dr3.gaiaedr3,
                WD_gaia_dr3.pwd,
                WD_gaia_dr3.gmag_vega,
            )
            .join(WD_gaia_dr3, on=(CatalogToGaia_DR3.target_id == WD_gaia_dr3.gaiaedr3))
            .where(
                WD_gaia_dr3.pwd > 0.5,
                WD_gaia_dr3.gmag_vega <= 20.0,
                CatalogToGaia_DR3.version_id == version_id,
                CatalogToGaia_DR3.best >> True,
            )
        )

        if query_region:
            query = query.join_from(CatalogToGaia_DR3, Catalog).where(
                peewee.fn.q3c_radial_query(
                    Catalog.ra,
                    Catalog.dec,
                    query_region[0],
                    query_region[1],
                    query_region[2],
                )
            )

        return query


class MWM_WD_Gaia_boss_Carton(BaseCarton):
    """MWM Gaia White Dwarfs.

    Definition:

    bp_rp=gaiadr3.phot_bp_mean_mag-gaiadr3.phot_rp_mean_mag
    par_log = log10(gaiadr3.parallax/100.)

    First select:
    gaiadr3.phot_g_mean_mag < 20.0
    and gaiadr3.parallax*power(10.,0.2*(22.-gaiadr3.phot_g_mean_mag))> 3.
    and phot_g_mean_mag+5*par_log-0.5-5.5*(bp_rp+0.5) > 0.
    and ((bp_rp<0.4) or ((gaiadr3.phot_g_mean_mag+5*par_log)> 11+2.2*(bp_rp-0.5) and (bp_rp >0.4))

    Then select:
    gaiadr3.phot_g_mean_mag+5*par_log - 6.7-2.2*bp_rp  > 0. or bp_rp < -0.6

    And finally
    gDR3spur_main.fidelity_v2 > 0.9

    """

    name = "mwm_wd_gaia_boss"
    mapper = "MWM"
    category = "science"
    program = "mwm_wd"
    instrument = "BOSS"
    cadence = "dark_2x2"
    priority = 1401
    can_offset = True

    def build_query(self, version_id, query_region=None):
        self.database.execute_sql("SET LOCAL random_page_cost = 0.1;")

        bp_rp = Gaia_DR3.phot_bp_mean_mag - Gaia_DR3.phot_rp_mean_mag
        gaiaG = Gaia_DR3.phot_g_mean_mag
        par_log = fn.log10(Gaia_DR3.parallax / 100.0)

        gaia_cut = (
            (gaiaG < 20.0)
            & (Gaia_DR3.parallax * fn.power(10.0, 0.2 * (22.0 - gaiaG)) > 3.0)
            & (gaiaG + 5 * par_log - 0.5 - 5.5 * (bp_rp + 0.5) > 0.0)
            & (
                (bp_rp < 0.4)
                | ((bp_rp > 0.4) & ((gaiaG + 5 * par_log) > (11 + 2.2 * (bp_rp - 0.5))))
            )
        )

        query = (
            CatalogToGaia_DR3.select(
                CatalogToGaia_DR3.catalogid,
                Gaia_DR3.source_id,
                Gaia_DR3.ra,
                Gaia_DR3.dec,
                bp_rp.alias("bp_rp"),
                gaiaG,
                Gaia_DR3.parallax,
                Gedr3spur_main.fidelity_v2,
            )
            .join(Gaia_DR3)
            .join(Gedr3spur_main, on=(Gaia_DR3.source_id == Gedr3spur_main.source_id))
            .where(
                CatalogToGaia_DR3.version_id == version_id,
                CatalogToGaia_DR3.best >> True,
                gaia_cut,
                ((gaiaG + 5 * par_log - 6.7 - 2.2 * bp_rp) > 0.0) | (bp_rp < -0.6),
                Gedr3spur_main.fidelity_v2 > 0.9,
            )
        )

        if query_region:
            query = query.join_from(CatalogToGaia_DR3, Catalog).where(
                peewee.fn.q3c_radial_query(
                    Catalog.ra,
                    Catalog.dec,
                    query_region[0],
                    query_region[1],
                    query_region[2],
                )
            )

        return query


class MWM_WD_PWD_boss_2x1_Carton(MWM_WD_PWD_boss_Carton):
    """
    Shorthand name:  mwm_wd_pwd_boss_2x1

    Existing carton code:
    Simplified Description of selection criteria :
    Unchanged from previous version of mwm_wd_pwd_boss
    Pseudo-code: ---
    Return columns: Same as previous
    Metadata:
    Priority: 1402
    Cadence:  dark_2x1
    Instrument: BOSS
    can_offset = True
    Lead contact: Nikola Gentile Fusillo
    """

    name = "mwm_wd_pwd_boss_2x1"
    mapper = "MWM"
    category = "science"
    program = "mwm_wd"
    instrument = "BOSS"
    cadence = "dark_2x1"
    priority = 1402
    can_offset = True


class MWM_WD_PWD_boss_1x3_Carton(MWM_WD_PWD_boss_Carton):
    """
    Shorthand name:  mwm_wd_pwd_boss_1x3

    Existing carton code:
    Simplified Description of selection criteria :
    Unchanged from previous version of mwm_wd_pwd_boss
    Pseudo-code: ---
    Return columns: Same as previous
    Metadata:
    Priority: 1404
    Cadence:  dark_1x3
    Instrument: BOSS
    can_offset = True
    Lead contact: Nikola Gentile Fusillo
    """

    name = "mwm_wd_pwd_boss_1x3"
    mapper = "MWM"
    category = "science"
    program = "mwm_wd"
    instrument = "BOSS"
    cadence = "dark_1x3"
    priority = 1404
    can_offset = True


class MWM_WD_PWD_boss_1x2_Carton(MWM_WD_PWD_boss_Carton):
    """
    Shorthand name:  mwm_wd_pwd_boss_1x2

    Existing carton code:
    Simplified Description of selection criteria :
    Unchanged from previous version of mwm_wd_pwd_boss
    Pseudo-code: ---
    Return columns: Same as previous
    Metadata:
    Priority: 1406
    Cadence:  dark_1x2
    Instrument: BOSS
    can_offset = True
    Lead contact: Nikola Gentile Fusillo
    """

    name = "mwm_wd_pwd_boss_1x2"
    mapper = "MWM"
    category = "science"
    program = "mwm_wd"
    instrument = "BOSS"
    cadence = "dark_1x2"
    priority = 1406
    can_offset = True


class MWM_WD_Gaia_boss_2x1_Carton(MWM_WD_Gaia_boss_Carton):
    """
    Shorthand name:  mwm_wd_gaia_boss_2x1

    Existing carton code:
    Simplified Description of selection criteria :
    Unchanged from previous version of mwm_wd_pwd_boss
    Pseudo-code: ---
    Return columns: Same as previous
    Metadata:
    Priority: 1403
    Cadence:  dark_2x1
    Instrument: BOSS
    can_offset = True
    Lead contact: Nikola Gentile Fusillo
    """

    name = "mwm_wd_gaia_boss_2x1"
    mapper = "MWM"
    category = "science"
    program = "mwm_wd"
    instrument = "BOSS"
    cadence = "dark_2x1"
    priority = 1403
    can_offset = True


class MWM_WD_Gaia_boss_1x3_Carton(MWM_WD_Gaia_boss_Carton):
    """
    Shorthand name:  mwm_wd_gaia_boss_1x3

    Existing carton code:
    Simplified Description of selection criteria :
    Unchanged from previous version of mwm_wd_pwd_boss
    Pseudo-code: ---
    Return columns: Same as previous
    Metadata:
    Priority: 1405
    Cadence:  dark_1x3
    Instrument: BOSS
    can_offset = True
    Lead contact: Nikola Gentile Fusillo
    """

    name = "mwm_wd_gaia_boss_1x3"
    mapper = "MWM"
    category = "science"
    program = "mwm_wd"
    instrument = "BOSS"
    cadence = "dark_1x3"
    priority = 1405
    can_offset = True


class MWM_WD_Gaia_boss_1x2_Carton(MWM_WD_Gaia_boss_Carton):
    """
    Shorthand name:  mwm_wd_gaia_boss_1x2

    Existing carton code:
    Simplified Description of selection criteria :
    Unchanged from previous version of mwm_wd_pwd_boss
    Pseudo-code: ---
    Return columns: Same as previous
    Metadata:
    Priority: 1407
    Cadence:  dark_1x2
    Instrument: BOSS
    can_offset = True
    Lead contact: Nikola Gentile Fusillo
    """

    name = "mwm_wd_gaia_boss_1x2"
    mapper = "MWM"
    category = "science"
    program = "mwm_wd"
    instrument = "BOSS"
    cadence = "dark_1x2"
    priority = 1407
    can_offset = True



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2020-07-01
# @Filename: mwm_tess_rgb.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

import numpy
import peewee
from peewee import fn

from sdssdb.peewee.sdss5db.catalogdb import (
    Catalog,
    CatalogToGaia_DR3,
    CatalogToTIC_v8,
    Gaia_DR3,
    TIC_v8,
)

from target_selection.cartons import BaseCarton


def h2exp(hmag, sn=100, exptime=15.0):
    """This function takes in a hmag and given signal to noise and spits back
    the required time. Based on Hmag = 11 at S/N 100 in an hour.
    """

    # Scale the hmag based on t = (1 hour)*10^(0.4*(H-11))
    # Then I cut it up into 15 minute exposures.
    time = 60 * (sn**2 / 100.0**2) * 10 ** (0.4 * (hmag - 11))
    nexp = numpy.array(numpy.round(time / exptime))

    # Min value is 1
    nexp[(nexp == 0)] = 1

    # Set Nan's to nan
    nexp[numpy.isnan(hmag)] = numpy.nan

    return nexp


class MWM_TESS_RGB_apogee_Carton(BaseCarton):
    """MWM TESS RGB apogee Carton.

    Definition:

    For v0.5
        - Jmag - Kmag > 0.5 (get red stars)
        - Hmag < 12 (get needed SNR w/APOGEE)
        - Tmag < 13 (get detections of oscillations) [Tmag := TESS magnitude]
        - M_H (absolute H band magnitude) < 1 (get giants)
            where: MH = Hmag - 10 + 5.0 * log10(parallax)
            where the parallax is in mas.

    For v1
    remove the |b| > 20 criterion
    change the J-K limit to 0.3
    change the MH, Tmag criteria as described below
    h2exp remains the same, yes!

    Simplified Description of selection criteria:

    Jmag - Kmag > 0.3,
    Hmag < 12,
    MH <= 3 where MH=Hmag-10+5.0*log10(parallax)
    where the parallax is in mas.

    if MH < 1 then Tmag must be  < 13 else Tmag must be < -5*MH+18

    Gaia DR2 parameters to be converted to Gaia DR3:
    use Gaia DR3 parallaxes instead of Gaia DR2

    use joins tic_v8 → catalog_to_tic_v8 → catalog → catalog_to_gaia_dr3 → gaia_dr3

    Return columns: Unchanged

    Metadata:
    priority=2820
    cadence as before
    can_offset=True
    instrument=APOGEE

    """

    name = "mwm_tess_rgb_apogee"  # there is an underscore between tess and rgb
    mapper = "MWM"
    category = "science"
    program = "mwm_tessrgb"  # there is no underscore between tess and rgb
    instrument = "APOGEE"
    cadence = None
    priority = 2820
    can_offset = True

    def build_query(self, version_id, query_region=None):
        MH = TIC_v8.hmag - 10 + 5 * fn.log(Gaia_DR3.parallax)

        # v0.5
        #         query = (TIC_v8
        #                  .select(CatalogToTIC_v8.catalogid,
        #                          TIC_v8.hmag,
        #                          TIC_v8.jmag,
        #                          TIC_v8.kmag,
        #                          TIC_v8.tmag,
        #                          TIC_v8.plx)
        #                  .join(CatalogToTIC_v8)
        #                  .where(CatalogToTIC_v8.version_id == version_id,
        #                         CatalogToTIC_v8.best >> True)
        #                  .where((TIC_v8.jmag - TIC_v8.kmag) > 0.5)
        #                  .where(((TIC_v8.plx > 0) & (MH < 1)) | (TIC_v8.plx < 0),
        #                         fn.abs(TIC_v8.gallat) > 20)
        #                  .where(TIC_v8.hmag < 12)
        #                  .where(TIC_v8.tmag < 13))

        query = (
            Catalog.select(
                CatalogToTIC_v8.catalogid,
                TIC_v8.id,
                TIC_v8.hmag,
                TIC_v8.jmag,
                TIC_v8.kmag,
                TIC_v8.tmag,
                Gaia_DR3.parallax,
            )
            .join(CatalogToTIC_v8, on=(Catalog.catalogid == CatalogToTIC_v8.catalogid))
            .join(TIC_v8, on=(CatalogToTIC_v8.target_id == TIC_v8.id))
            .switch(Catalog)
            .join(CatalogToGaia_DR3, on=(Catalog.catalogid == CatalogToGaia_DR3.catalogid))
            .join(Gaia_DR3, on=(CatalogToGaia_DR3.target_id == Gaia_DR3.source_id))
            .where(
                CatalogToTIC_v8.version_id == version_id,
                CatalogToTIC_v8.best >> True,
                CatalogToGaia_DR3.best >> True,
                (TIC_v8.jmag - TIC_v8.kmag) > 0.3,
                TIC_v8.hmag < 12,
                Gaia_DR3.parallax > 0,
                ((MH <= 1) & (TIC_v8.tmag < 13))
                | ((MH > 1) & (MH <= 3) & (TIC_v8.tmag < (-5 * MH + 18))),
            )
        )

        if query_region:
            query = query.join_from(CatalogToTIC_v8, Catalog).where(
                peewee.fn.q3c_radial_query(
                    Catalog.ra,
                    Catalog.dec,
                    query_region[0],
                    query_region[1],
                    query_region[2],
                )
            )

        return query

    def post_process(self, model, **kwargs):
        data = numpy.array(
            model.select(model.catalogid, model.hmag).tuples(),
            dtype=[("catalogid", numpy.int64), ("hmag", numpy.float32)],
        )
        n_exp = h2exp(data["hmag"], sn=80)

        # Below old code is for historical reference.
        # Below values1 is a generator.
        # values1 = ((int(data['catalogid'][ii]),
        #          'bright_flexible_' + str(int(n_exp[ii])) + 'x1'
        #           if not numpy.isnan(n_exp[ii]) else None)
        #          for ii in range(len(data)))

        # We use name 'values1' instead of 'values' since values()
        # is a Python built-in function.
        #
        # For int(n_exp[ii]) == 1, we use bright_1x1 since
        # there is no cadence bright_flexible_1x1
        #
        # Below values1 is a list and then later we convert it to a tuple.
        values1 = [None] * len(data)
        for ii in range(len(data)):
            if not numpy.isnan(n_exp[ii]):
                if int(n_exp[ii]) == 1:
                    current_cadence = "bright_1x1"
                else:
                    current_cadence = "bright_flexible_" + str(int(n_exp[ii])) + "x1"
                values1[ii] = (int(data["catalogid"][ii]), current_cadence)
            else:
                values1[ii] = None

        values1 = tuple(values1)
        vl = peewee.ValuesList(values1, columns=("catalogid", "cadence"), alias="vl")

        (
            model.update(cadence=vl.c.cadence).from_(vl).where(model.catalogid == vl.c.catalogid)
        ).execute()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2020-05-26
# @Filename: __main__.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

import importlib
import logging
import os
import sys
import time
import warnings
from glob import glob

import click
import peewee

import target_selection as tsmod
from target_selection.exceptions import TargetSelectionError, TargetSelectionUserWarning


def all_subclasses(cls):
    return set(cls.__subclasses__()).union(
        [s for c in cls.__subclasses__() for s in all_subclasses(c)]
    )


def connect(profile=None, dbname=None, user=None, host=None, port=None):
    """Connects the database."""

    from sdssdb.peewee.sdss5db.catalogdb import database

    if profile:
        database.set_profile(profile)

    if dbname or user or host or port:
        database.connect(dbname=dbname, user=user, host=host, port=port)

    return database


@click.group()
@click.option("--profile", "-p", type=str, default=None)
@click.option("--dbname", "-d", type=str, default=None)
@click.option("--user", "-u", type=str, default=None)
@click.option("--host", "-h", type=str, default=None)
@click.option("--port", "-P", type=int, default=None)
@click.option("--verbose", "-v", is_flag=True, help="outputs extra debug information")
@click.option("--save-log", type=str, default=None, help="saves the log to a file.")
@click.pass_context
def target_selection(ctx, profile, dbname, user, host, port, verbose, save_log):
    """Performs tasks related to target selection for SDSS-V."""

    if verbose:
        tsmod.log.set_level(logging.DEBUG)

    if save_log:
        tsmod.log.start_file_logger(os.path.realpath(save_log), mode="a", rotating=False)

    database = connect(profile, dbname, user, host, port)
    if not database.connected:
        raise TargetSelectionError("database is not connected.")

    ctx.obj = {"database": database}


@target_selection.command()
@click.argument("TARGETING-PLAN", nargs=1, type=str)
@click.option(
    "--config-file",
    type=click.Path(exists=True, dir_okay=False),
    help="the file to read. Defaults to the internal configuration file.",
)
@click.option(
    "--overwrite",
    is_flag=True,
    help="drop intermediate tables if they exist",
)
@click.option(
    "--keep",
    is_flag=True,
    help="keep intermediate tables after loading",
)
@click.option(
    "--skip-query",
    is_flag=True,
    help="do not run the query, only load intermediate table if it exists",
)
@click.option(
    "--region",
    "-r",
    type=float,
    nargs=3,
    default=None,
    help="the region (ra, dec, radius) to query",
)
@click.option(
    "--limit",
    "-l",
    type=int,
    default=None,
    help="limit number of targets in the carton",
)
@click.option(
    "--load/--no-load",
    is_flag=True,
    default=True,
    help="whether to load data into targetdb.target",
)
@click.option(
    "--include",
    "-i",
    type=str,
    help="comma-separated carton names to include",
)
@click.option(
    "--exclude",
    "-e",
    type=str,
    help="comma-separated carton names to exclude",
)
@click.option(
    "--write-table",
    "-w",
    is_flag=True,
    help="write table of loaded targets as a FITS file",
)
@click.option(
    "--exclude-open-fiber",
    is_flag=True,
    help="do not process open fiber cartons",
)
@click.option(
    "--add-magnitudes/--no-add-magnitudes",
    "-M/-m",
    default=True,
    help="Whether to add optical magnitudes during the carton run.",
)
def run(
    targeting_plan,
    config_file,
    overwrite,
    keep,
    region,
    load,
    skip_query,
    include,
    exclude,
    write_table,
    limit,
    exclude_open_fiber,
    add_magnitudes,
):
    """Runs target selection for all cartons."""

    from target_selection.cartons import BaseCarton
    from target_selection.cartons.tools import get_file_carton

    # Reload the cartons now that we have a connection.
    importlib.reload(sys.modules["target_selection.cartons"])

    carton_classes = {Carton.name: Carton for Carton in all_subclasses(BaseCarton)}

    if len(carton_classes) == 0:
        raise TargetSelectionError("no carton classes found.")

    try:
        config_plan = tsmod.config[targeting_plan]
    except KeyError:
        raise TargetSelectionError(f"cannot find configuration for plan {targeting_plan}.")

    carton_names = config_plan["cartons"]

    if exclude:
        carton_names = [cn for cn in carton_names if cn not in exclude.split(",")]

    if include:
        carton_names = include.split(",")
        for c in carton_names:
            if c not in carton_classes:
                raise ValueError(f"Carton {c} does not exist.")

    Cartons = []
    if carton_names is not None:
        Cartons = [carton_classes[carton_name] for carton_name in carton_names]
    else:
        warnings.warn(
            "There are no regular cartons in target_selection.yml.",
            TargetSelectionUserWarning,
        )

    if not exclude_open_fiber:
        if "open_fiber_path" not in config_plan:
            warnings.warn(
                'Plan does not specify an "open_fiber_path".',
                TargetSelectionUserWarning,
            )
        else:
            open_fiber_path = os.path.expandvars(config_plan["open_fiber_path"])
            if not os.path.exists(open_fiber_path):
                warnings.warn(
                    f"Open fiber path {open_fiber_path} does not exist.",
                    TargetSelectionUserWarning,
                )
            else:
                open_fiber_file_list_path = os.path.join(
                    open_fiber_path, "open_fiber_file_list.txt"
                )
                if not os.path.exists(open_fiber_file_list_path):
                    tsmod.log.info("open_fiber_path = " + open_fiber_path)
                    open_fiber_files = glob(os.path.join(open_fiber_path, "*.fits"))
                else:
                    tsmod.log.info("open_fiber file_list_path = " + open_fiber_file_list_path)
                    flist = open(open_fiber_file_list_path, "r")
                    open_fiber_files = []
                    for fline in flist:
                        open_fiber_files.append(os.path.join(open_fiber_path, fline.strip()))
                    flist.close()

                # We sort open_fiber_files so that the files are processed
                # in the same order every time.
                open_fiber_files.sort()

                # historical code
                # base_open_fiber_files = [None] * len(open_fiber_files)
                # for i in range(len(open_fiber_files)):
                #    tokens = open_fiber_files[i].split('/')
                #    base_open_fiber_files[i] = tokens[-1]
                #    tokens = base_open_fiber_files[i].split('.')
                #    base_open_fiber_files[i] = tokens[0].lower()

                OpenFiberCartons = [None] * len(open_fiber_files)
                for i in range(len(open_fiber_files)):
                    OpenFiberCartons[i] = get_file_carton(open_fiber_files[i])

                if include:
                    OpenFiberCartons = [OFC for OFC in OpenFiberCartons if OFC.name in include]
                Cartons += OpenFiberCartons

                tsmod.log.info(f"{len(OpenFiberCartons)} open fiber cartons selected.")
    else:
        tsmod.log.info("Excluding open fiber cartons.")

    for Carton in Cartons:
        carton = Carton(targeting_plan, config_file=config_file)

        tsmod.log.header = f"({carton.name}): "
        tsmod.log.info(f"Running target selection for carton {carton.name!r}.")

        if carton.check_targets() and not overwrite:
            raise ValueError(
                f"Found existing targets for carton {carton.name!r} with plan {carton.plan!r}."
            )

        if not skip_query:
            carton.run(
                query_region=(region or None),
                overwrite=overwrite,
                limit=limit,
                add_optical_magnitudes=add_magnitudes,
            )
            if write_table:
                carton.write_table()
        else:
            tsmod.log.debug("Skipping query.")

        if load:
            time.sleep(60)
            carton.load(overwrite=overwrite)
        else:
            tsmod.log.info("Not loading data into targetdb.target.")

        if load and write_table:
            carton.write_table(mode="targetdb")

        if not keep:
            tsmod.log.info(f"Dropping temporary table {carton.path!r}.")
            carton.drop_table()

    tsmod.log.header = ""


@target_selection.command("load-files")
@click.option("--program", type=str, default="open_fiber", show_default=True)
@click.option("--category", type=str, default="open_fiber", show_default=True)
@click.argument("PLAN", type=str, nargs=1)
@click.argument("FILES", nargs=-1, required=True)
def load_files(plan, files, program, category):
    """Loads a carton from a FITS file."""

    from target_selection.cartons.tools import get_file_carton

    for fn in files:
        Carton = get_file_carton(fn, "", category, program)
        carton = Carton(plan)
        carton.run(overwrite=True)
        carton.load(overwrite=True)


@target_selection.command()
@click.argument("TARGETING-PLAN", type=str)
@click.option("--tables", is_flag=True, help="also remove intermediate tables")
def clear(targeting_plan, tables):
    """Clear all data for a target selection plan."""

    from sdssdb.peewee.sdss5db import targetdb as tdb

    from target_selection.cartons import BaseCarton

    if tables:
        # Cartons = BaseCarton.__subclasses__()
        # find grandchildren classes as well
        Cartons = all_subclasses(BaseCarton)
        for Carton in Cartons:
            Carton(targeting_plan).drop_table()

    try:
        tdb.Version.get(plan=targeting_plan, target_selection=True).delete_instance()
    except peewee.DoesNotExist:
        pass


@target_selection.command()
@click.argument("XMATCH-PLAN", type=str)
@click.option(
    "--file",
    type=click.Path(exists=True, dir_okay=False),
    help="the file to read. Defaults to the internal configuration file.",
)
@click.pass_context
def xmatch(ctx, xmatch_plan, file):
    """Runs catalogue cross-matching from a configuration file."""

    from target_selection.xmatch import XMatchPlanner

    database = ctx.obj["database"]
    xmatch = XMatchPlanner.read(database, xmatch_plan, config_file=file)
    xmatch.run()


@target_selection.command()
@click.option("--cpus", type=int, default=1, help="Number of CPUs to use.")
@click.pass_context
def skies(ctx, cpus=1):
    """Runs the parent sky catalogue creation routine."""

    from target_selection.skies import create_sky_catalogue

    database = ctx.obj["database"]
    create_sky_catalogue(database, n_cpus=cpus)


if __name__ == "__main__":
    target_selection(obj={})



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2020-04-06
# @Filename: xmatch.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

import copy
import hashlib
import inspect
import os
import re
import types
import uuid
import warnings

import networkx
import numpy
import peewee
import rich.markup
import yaml
from networkx.algorithms import shortest_path
from peewee import SQL, Case, Model, fn

from sdssdb.connection import PeeweeDatabaseConnection
from sdssdb.utils.internals import get_row_count
from sdsstools import merge_config
from sdsstools._vendor.color_print import color_text

import target_selection
from target_selection.exceptions import (
    TargetSelectionNotImplemented,
    TargetSelectionUserWarning,
    XMatchError,
)
from target_selection.utils import (
    Timer,
    get_configuration_values,
    get_epoch,
    is_view,
    sql_apply_pm,
    vacuum_outputs,
    vacuum_table,
)


EPOCH = 2016.0
QUERY_RADIUS = 1.0

#: Reserve last 11 bits for the run id.
RUN_ID_BIT_SHIFT = 64 - 11

TEMP_SCHEMA = "sandbox"


class Version(peewee.Model):
    """Model for the version table."""

    id = peewee.AutoField()
    plan = peewee.TextField()
    tag = peewee.TextField()

    class Meta:
        table_name = "version"


class Catalog(peewee.Model):
    """Model for the output table."""

    catalogid = peewee.BigIntegerField(index=True, null=False)
    iauname = peewee.TextField(null=True)
    ra = peewee.DoubleField(null=False)
    dec = peewee.DoubleField(null=False)
    pmra = peewee.FloatField(null=True)
    pmdec = peewee.FloatField(null=True)
    parallax = peewee.FloatField(null=True)
    lead = peewee.TextField(null=False)
    version_id = peewee.IntegerField(null=False, index=True)


class TempCatalog(Catalog):
    """Temporary output table."""

    catalogid = peewee.BigIntegerField(index=True, null=False)
    version_id = peewee.IntegerField(index=False)

    class Meta:
        primary_key = False


def XMatchModel(
    Model,
    resolution=None,
    ra_column=None,
    dec_column=None,
    pmra_column=None,
    pmdec_column=None,
    is_pmra_cos=True,
    parallax_column=None,
    epoch_column=None,
    epoch=None,
    epoch_format="jyear",
    relational_table=None,
    has_duplicates=False,
    has_missing_coordinates=False,
    skip=False,
    skip_phases=None,
    query_radius=None,
    join_weight=1,
    database_options=None,
):
    """Expands the model `peewee:Metadata` with cross-matching parameters.

    The parameters defined can be accessed with the same name as
    ``Model._meta.xmatch.<parameter>`` (e.g.,
    ``Model._meta.xmatch.has_duplicates``).

    Parameters
    ----------
    resolution : float
        The spatial resolution of the catalogue, in arcsec.
    ra_column : str
        The name of the right ascension column. If not provided, an attempt
        will be made to recover it from the Q3C index, if it exists. Assumed
        to be in degrees.
    dec_column : str
        As ``ra``, for the declination column.
    pmra_column : str
        The RA proper motion column, assumed to be in milliarcseconds per year.
    pmdec_column : str
        As ``pmra_column`` for the declination proper motion.
    is_pmra_cos : bool
        Whether ``pmra_column`` provides the RA proper motion corrected from
        declination(``pmra * cos(dec)``) or not.
    parallax_column : str
        The column containing the parallax, assumed to be in arcsec.
    epoch_column : str
        The column containing the epoch of the target coordinates.
    epoch : float
        The epoch of the targets that applies to all the records in the
        table. ``epoch`` and ``epoch_column`` are mutually exclusive. If
        neither ``epoch_column`` nor ``epoch`` are defined, assumes that the
        epoch is ``2015.5``.
    epoch_format : str
        The format of the epoch. Either Julian year (``'jyear'``) or Julian
        date (``'jd'``).
    table_name : str
        Overrides the default model table name. This can be useful sometimes
        if, for example, a view has been created that contains only the
        columns from the main table needed for cross-matching.
    has_duplicates : bool
        Whether the table contains duplicates.
    has_missing_coordinates : bool
        Whether the catalogue contains rows in which the RA/Dec are null.
    skip : bool
        If `True`, the table will be used as a join node but will not be
        cross-matched. This is useful for testing and also if the table is
        in a previous version of the configuration file and you are using the
        ``base_version`` option but want to remove that table. It can also
        be used when setting a ``join_path`` for a model but otherwise don't
        want the table to be processed.
    skip_phases : list
        A list of cross-matching phases to be skipped for this model. Refer
        to the `.XMatchPlanner` documentation for definitions on what each
        phase does.
    query_radius : float
        The radius, in arcsec, to use in the radial query for cross-matching.
        If not provided defaults to the `.XMatchPlanner` value.
    join_weight : float
        The weight used by `.XMatchPlanner.get_join_paths` to determine
        the cost of using this table as a join. Lower weights translate to
        better chances of that join path to be selected.
    database_options : dict
        A dictionary of database configuration parameters to be set locally
        for this model for each processing phase transaction, temporarily
        overriding the default database configuration. Keys must be the
        database parameter to modify. The value can be a simple string with
        the value to set, or a dictionary that more accurately defines
        when the parameter will be applied. See `.XMatchPlanner` for more
        information.

    Returns
    -------
    :obj:`peewee:Model`
        The same input model with the additional cross-matching parameters
        added to the metadata namespace ``xmatch``.

    """

    meta = Model._meta
    meta.xmatch = types.SimpleNamespace()

    meta.xmatch.resolution = resolution or numpy.nan

    if not ra_column or not dec_column:
        indexes = meta.database.get_indexes(meta.table_name, meta.schema)
        for index in indexes:
            if "q3c" in index.sql.lower():
                match = re.match(r'.+q3c_ang2ipix\("*(\w+)"*, "*(\w+)"*\).+', index.sql)
                if match:
                    ra_column, dec_column = match.groups()

    meta.xmatch.ra_column = ra_column
    meta.xmatch.dec_column = dec_column

    meta.xmatch.pmra_column = pmra_column
    meta.xmatch.pmdec_column = pmdec_column
    meta.xmatch.is_pmra_cos = is_pmra_cos
    meta.xmatch.parallax_column = parallax_column

    assert ((epoch is None) & (epoch_column is None)) or (
        (epoch is not None) ^ (epoch_column is not None)
    ), "epoch and epoch_column are mutually exclusive."

    meta.xmatch.epoch = epoch
    meta.xmatch.epoch_column = epoch_column
    meta.xmatch.epoch_format = epoch_format

    meta.xmatch.relational_table = relational_table

    meta.xmatch.has_duplicates = has_duplicates
    meta.xmatch.has_missing_coordinates = has_missing_coordinates
    meta.xmatch.skip = skip
    meta.xmatch.skip_phases = skip_phases or []
    meta.xmatch.query_radius = query_radius

    meta.xmatch.row_count = int(
        get_row_count(
            meta.database,
            meta.table_name,
            schema=meta.schema,
            approximate=True,
        )
    )

    meta.xmatch.join_weight = join_weight

    meta.xmatch.database_options = database_options or {}

    return Model


class XMatchPlanner(object):
    """Prepares and runs catalogue cross-matching.

    This class prepares the execution of a cross-matching between multiple
    catalogues, the result being an output table of unique targets linked to
    each input catalog via a relational table. Target coordinates are
    propagated to a common epoch when proper motions are available.
    Instantiating the class only prepares the process and sets up the
    processing order; cross-matching itself happens by calling the `.run`
    method.

    The output table contains a sequential integer identifier for each unique
    target (``catalogid``), along with the following columns: ``ra``, ``dec``,
    ``pmra``, ``pmdec``, ``parallax``, ``lead``, and ``version_id``.
    ``version_id`` relates the record in the ``version`` table which contains
    the cross-matching plan and the tag of the code used when it was
    run. ``lead`` indicates from which one of the input catalogues the
    coordinates were obtained.

    The output table is related to each input catalogue via a many-to-many
    table with the format ``<output_table>_to_<catalogue_table>`` where
    ``<output_table>`` is the table name of the output table and
    ``<catalog_table>`` is the table name of the referred catalogue. Each
    relational table contains the unique identifier column, ``catalogid``,
    ``target_id`` pointing to the primary key of the referred catalogue,
    ``version_id``, and a boolean ``best`` indicating whether it is the best
    (closest) possible match when a target in the output table correspond to
    multiple targets in the input catalogue.

    The cross-matching process roughly follows the following process:

    |

    .. image:: _static/Catalogdb_Crossmatch.png
        :scale: 75 %
        :align: center

    |

    In practice the cross-matching process begins by creating a graph of all
    nodes (tables to be processed and additional tables, ``extra_nodes``, in
    the database) and edges (foreign keys relating two tables). This graph is
    used to determine join conditions and to establish the order in which the
    input models will be processed. The processing order is determined by the
    ``order`` and ``key`` input parameters. When ``key='row_count'``, tables
    are sorted by number of decreasing rows so that tables with more targets
    are processed first (note that to speed things up the row count is always
    the latest known approximate determined by ``ANALYZE``); if
    ``key='resolution'`` the associated spatial resolution for a catalogue is
    used to process catalogues with high resolution first. If
    ``order='hierarchical'``, all the tables are divided into as many
    disconnected subgraphs as exist; then for each subgraph the maximum row
    count or minim resolution is calculated (depending on the value of
    ``key``). Subgraphs are sorted based on this result and then tables
    belonging to each subgraph are sorted by key. If ``order='global'`` the
    ``key`` ordering is applied to all tables without taking into account
    subgraphs.

    To speed things up unique targets are initially inserted into a temporary
    table ``<output_table>_<uid>`` where ``<output_table>`` is the name of
    the output table and ``<uid>`` is a unique identifier based on the version.

    Once the order has been determined and when `.run` is called, each table
    model is processed in order. The first model is just ingested completely
    into the temporary table and its associated relational table is created if
    it does not exist (the equivalent of *phase 3* below).

    For each additional model the following three stages are applied:

    - In *phase 1* we determine what targets in the input model have an
      existing cross-match to targets already ingested into the temporary
      table. To do that we build all possible joins between the model and the
      temporary table. If multiple joins are possible via a given table only
      the shortest is used (see `.get_join_paths`). For all matched targets we
      insert entries in the relational table. The *lead* of the original
      entries is not changed.

    - In *phase 2* we perform the actual cross-match between targets in the
      temporary table and the ones in the input catalogue. Currently the only
      cross-matching method available is a spatial cone query with radius
      ``query_radius``. All matched targets are added to the relational table
      and the one with the smallest distance is defined as *best*.

    - In *phase 3* we determine any target in the input catalogue that has not
      been cross-matched at this point and insert them into the temporary table
      as new entries. The *lead* is set to the input catalogue and the
      one-to-one match is added to the relational table.

    In phases 1 and 3 the queries are initially stored as a Postgresql
    temporary table for efficiency, and then copied to the relational table.
    After all the tables have been processed the output temporary table is
    inserted in bulk into the output table and dropped.

    In addition to the limitations of the spatial cone query method, the
    following caveats are known:

    - Input tables with duplicate targets are not currently supported.

    - In *phase 2* there is no current measure in place for the same target
      to be associated with more than one catalogid (i.e., each cross-match
      is performed independently).

    Parameters
    ----------
    database : peewee:PostgresqlDatabase
        A `peewee:PostgresqlDatabase` to the database the tables to
        cross-match.
    models : list
        The list of `.XMatchModel` classes to be cross-matched. If the model
        correspond to a non-existing table it will be silently ignored.
    plan : str
        The cross-matching plan version.
    run_id : int
        An integer to identify this run of cross-matching. The ID is bit
        shifted `.RUN_ID_BIT_SHIFT` positions and added to the catalogid.
        This allows to quickly associate a ``catalogid`` with a run without
        having to query ``catalogdb``. A ``run_id`` cannot be used if there
        are targets already in ``catalog`` using that same ``run_id``.
    version_id
        The ``catalogdb.version.id`` cross-match version to use. Normally this
        will be `None`, in which case a new version id will be created. For
        "addendum" runs, this should be set to the run to which to append.
    extra_nodes : list
        List of PeeWee models to be used as extra nodes for joins (i.e.,
        already established cross-matches between catalogues). This models
        are not processed or inserted into the output table.
    order : str or list
        The type of sorting to be applies to the input models to decide in
        what order to process them. Currently allowed values are
        ``'hierarchical'`` and ``'global'`` (refer to the description above).
        The order can also be a list of table names, in which case that order
        is used without any further sorting.
    key : str
        The key to be used while sorting. Can be ``'row_count'`` or
        ``'resolution'``.
    epoch : float
        The epoch to which to convert all the target coordinates as they are
        inserted into the output table.
    start_node : str
        If specified, the name of the table that will be inserted first
        regarding of the above sorting process.
    query_radius : float
        The radius, in arcsec, for cross-matching between existing targets.
        Used in phase 2. Defaults to 1 arcsec.
    schema : str
        The schema in which all the tables to cross-match live (multiple
        schemas are not supported), and the schema in which the output tables
        will be created.
    output_table : str
        The name of the output table. Defaults to ``catalog``.
    temp_schema
        The schema where the temporary ``catalog`` table will be initially created.
    log
        A logger to which to log messages. If not provided the ``target_selection``
        logger is used.
    log_path : str
        The path to which to log or `False` to disable file logging.
    debug : bool or int
        Controls the level to which to log to the screen. If `False` no logging
        is done to ``stdout``. If `True` the logging level is set to ``debug``
        (all messages). It's also possible specify a numerical value for the
        `logging level <logging>`.
    show_sql : bool
        Whether to log the full SQL queries being run.
    sample_region : tuple
        Allows to specify a 3-element tuple with the
        (``ra``, ``dec``, ``radius``) of the region to which to limit the
        cross-match. All values must be in degrees. It can also be a list
        of tuples, in which case the union of all the regions will be sampled.
    path_mode : str
        The mode to look for join paths to link tables to output catalog.
        ``original`` mode uses a scorched earth algorithm that iteratively
        looks the shortest path and removes the first node trasspased and
        starts again until no shortest paths are available. ``full`` mode
        retrieves all the paths that are not a subsample of another path.
        ``config_list`` mode takes the list of paths from ``join_paths``
        parameter in the .yml configuration file.
    join_paths : list
        When using path_mode=``config_list`` is the list of paths to link
        tables to output catalog table in phase_1.
    database_options : dict
        A dictionary of database configuration parameters to be set locally
        during each phase transaction, temporarily overriding the default
        database configuration. Keys must be the database parameter to modify.
        The value can be a simple string with the value to set, or a dictionary
        that more accurately defines when the parameter will be applied. For
        example ::

            database_options:
                work_mem: '2GB'
                temp_buffers : {value: '500MB', phases: [3]}

        In this case the ``temp_buffers='500MB'`` option will only be set for
        phase 3. Configuration options to be used for a specific table can
        be set up when defining the `.XMatchModel`.

    """

    def __init__(
        self,
        database,
        models,
        plan,
        run_id,
        version_id=None,
        extra_nodes=[],
        order="hierarchical",
        key="row_count",
        epoch=EPOCH,
        start_node=None,
        query_radius=None,
        schema="catalogdb",
        temp_schema=TEMP_SCHEMA,
        output_table="catalog",
        log=None,
        log_path="./xmatch_{plan}.log",
        debug=False,
        show_sql=False,
        sample_region=None,
        database_options=None,
        path_mode="full",
        join_paths=None,
    ):
        self.log = log or target_selection.log
        self.log.header = ""

        if log_path:
            log_path = os.path.realpath(log_path)
            if self.log.fh:
                self.log.removeHandler(self.log.fh)
                self.log.fh = None
            self.log.start_file_logger(log_path.format(plan=plan), rotating=False, mode="a")

        if debug is True:
            self.log.sh.setLevel(0)
        elif debug is False:
            self.log.sh.setLevel(100)
        else:
            self.log.sh.setLevel(debug)

        self.schema = schema
        self.temp_schema = temp_schema

        self.output_table = output_table
        self.md5 = hashlib.md5(plan.encode()).hexdigest()[0:16]
        self._temp_table = self.output_table + "_" + self.md5

        self.database = database
        assert self.database.connected, "database is not connected."

        self.plan = plan
        self.run_id = run_id

        self.version_id = version_id
        self.is_addendum = version_id is not None

        self.tag = target_selection.__version__
        self.log.info(f"plan = {self.plan!r}; run_id = {self.run_id}; tag = {self.tag!r}.")
        self.log.info(f"Reference Epoch = {epoch}")
        self.models = {model._meta.table_name: model for model in models}
        self.extra_nodes = {model._meta.table_name: model for model in extra_nodes}
        self._check_models()

        self._options = {
            "query_radius": query_radius or QUERY_RADIUS,
            "show_sql": show_sql,
            "sample_region": sample_region,
            "epoch": epoch,
            "database_options": database_options or None,
        }

        qrd = self._options["query_radius"]
        self.log.info(f"Query radius = {qrd}")

        if self._options["sample_region"]:
            sample_region = self._options["sample_region"]
            self.log.warning(f"Using sample region {sample_region!r}.")

        self._log_db_configuration()

        self.model_graph = None
        self.update_model_graph()

        self.process_order = []
        self.set_process_order(order=order, key=key, start_node=start_node)

        self._max_cid = self.run_id << RUN_ID_BIT_SHIFT
        self.path_mode = path_mode

        if path_mode == "config_list":
            join_paths_warning = "join_paths needed for path_mode=config_list"
            assert join_paths is not None, join_paths_warning
            self.join_paths = join_paths

    @classmethod
    def read(cls, in_models, plan, config_file=None, **kwargs):
        """Instantiates `.XMatchPlanner` from a configuration file.

        The YAML configuration file must organised by plan string (multiple
        plans can live in the same file). Any parameter that
        `.XMatchPlanner` accepts can be passed via the configuration file.
        Additionally, the configuration file accepts the two extra parameters:
        ``exclude``, a list of table names that will be ignored (this is useful
        if you pass a datbase or base class as ``in_models`` and want to ignore
        some of the models), and ``tables``, a dictionary of table names with
        parameters to be passed to `.XMatchModel` for its corresponding model.
        An example of a valid configuration files is:

        .. code-block:: yaml

            '0.1.0':
                order: hierarchical
                key: resolution
                query_radius: 1.
                schema: catalogdb
                output_table: catalog
                start_node: tic_v8
                debug: true
                log_path: false
                exclude: ['catwise']
                tables:
                    tic_v8:
                        ra_column: ra
                        dec_column: dec
                        pmra_column: pmra
                        pmdec_column: pmdec
                        is_pmra_cos: true
                        parallax_column: plx
                        epoch: 2015.5
                    gaia_dr2_source:
                        ra_column: ra
                        dec_column: dec
                        pmra_column: pmra
                        pmdec_column: pmdec
                        is_pmra_cos: true
                        parallax_column: parallax
                        epoch: 2015.5
                        skip: true

        It is also possible to use a parameter ``base_plan`` pointing to a
        previous plan string. In that case the previous plan configuration
        will be used as base and the new values will be merged (the update
        happens recursively as with normal Python dictionaries).

        Note that only models that match the table names in ``tables`` will be
        passed to `.XMatchPlanner` to be processed; any other table will be
        used as an extra joining node unless it's listed in ``exclude``, in
        which case it will be ignored completely. It's possible to set the
        ``skip`` option for a table; this has the same effect as removing the
        entry in ``table``.

        Parameters
        ----------
        in_models
            The models to cross-match. Can be a list or tuple of PeeWee
            :obj:`peewee:Model` instances, a base class from which all the
            models to use subclass, or a
            `~sdssdb.connection.PeeweeDatabaseConnection` to the database
            containing the models. In the latter case, the models must have
            been imported so that they are available via the ``models``
            attribute.
        plan : str
            The cross-matching plan.
        config_file : str or dict
            The path to the configuration file to use. Defaults to
            ``config/xmatch.yml``. The file must contain a hash with the
            cross-match plan.
        kwargs : dict
            User arguments that will override the configuration file values.

        """

        # HACK: this ensures that the catalogdb.models are populated. In principle thos would not
        # work if schema != catalogdb but anyway many other things would fail in that case.
        from sdssdb.peewee.sdss5db import catalogdb  # noqa

        if isinstance(in_models, (list, tuple)):
            models = in_models
        elif inspect.isclass(in_models) and issubclass(in_models, Model):
            database = in_models._meta.database
            models = set(in_models.__subclasses__())
            while True:
                old_models = models.copy()
                for model in old_models:
                    models |= set(model.__subclasses__())
                if models == old_models:
                    break
        elif isinstance(in_models, PeeweeDatabaseConnection):
            database = in_models
            models = database.models.values()
        else:
            raise TypeError(f"invalid input of type {type(in_models)!r}")

        assert database.connected, "database is not connected."

        if config_file is None:
            config_file = os.path.dirname(target_selection.__file__) + "/config/xmatch.yml"

        config = XMatchPlanner._read_config(config_file, plan)
        table_config = config.pop("tables", {}) or {}
        exclude_nodes = config.pop("exclude_nodes", []) or []
        assert "schema" in config, "schema is required in configuration."
        schema = config["schema"]

        models = {
            model._meta.table_name: model for model in models if model._meta.schema == schema
        }

        xmatch_models = {}
        for table_name in table_config:
            if table_name not in models:
                continue
            table_params = table_config[table_name] or {}
            xmatch_models[table_name] = XMatchModel(models[table_name], **table_params)

        extra_nodes_config = config.pop("extra_nodes", [])
        for table_name in extra_nodes_config:
            if table_name in models:
                xmatch_models[table_name] = XMatchModel(models[table_name])
            else:
                warnings.warn(
                    f"Cannot find model for extra node {table_name!r}.",
                    TargetSelectionUserWarning,
                )

        extra_nodes = [
            models[table_name]
            for table_name in models
            if table_name not in xmatch_models and table_name not in exclude_nodes
        ]

        config.update(kwargs)

        signature = inspect.signature(XMatchPlanner)

        valid_kw = {}
        for kw in config:
            if kw not in signature.parameters:
                warnings.warn(
                    f"ignoring invalid configuration value {kw!r}.",
                    TargetSelectionUserWarning,
                )
                continue
            valid_kw[kw] = config[kw]

        return cls(database, xmatch_models.values(), plan, extra_nodes=extra_nodes, **valid_kw)

    @staticmethod
    def _read_config(file_, plan):
        """Reads the configuration file, recursively."""

        if isinstance(file_, dict):
            config = copy.deepcopy(file_)
        else:
            config = yaml.load(open(file_, "r"), Loader=yaml.SafeLoader)

        assert plan in config, f"plan {plan!r} not found in configuration."

        base_plan = config[plan].pop("base_plan", None)
        if base_plan:
            config = merge_config(config[plan], XMatchPlanner._read_config(file_, base_plan))
        else:
            config = config[plan]

        return config

    def _check_models(self):
        """Checks the input analyse models."""

        catalog_tname = self.output_table

        # Remove extra nodes that are relational tables because we'll use
        # temporary relational tables instead.
        self.extra_nodes = {
            tname: self.extra_nodes[tname]
            for tname in self.extra_nodes
            if (
                not tname.startswith(catalog_tname + "_to_")
                and not tname == catalog_tname
                and self.extra_nodes[tname].table_exists()
            )
        }

        for tname in list(self.models.keys()):
            model = self.models[tname]
            meta = model._meta

            view_exists = any(
                [is_view(self.database, tname, self.schema, mat) for mat in [True, False]]
            )

            if not model.table_exists() and not view_exists:
                self.log.warning(f"table {tname!r} does not exist.")
            elif tname == catalog_tname:
                pass
            elif tname.startswith(catalog_tname + "_to_"):
                pass
            elif meta.xmatch.skip:
                self.extra_nodes[tname] = model
            else:
                continue

            self.models.pop(tname)

        if len(self.models) == 0:
            raise XMatchError("no models to cross-match.")

    def _log_db_configuration(self):
        """Logs some key database configuration parameters."""

        parameters = [
            "shared_buffers",
            "effective_cache_size",
            "wal_buffers",
            "effective_io_concurrency",
            "work_mem",
            "max_worker_processes",
            "random_page_cost",
            "seq_page_cost",
            "cpu_index_tuple_cost",
            "cpu_operator_cost",
            "default_statistics_target",
            "temp_buffers",
            "plan_cache_mode",
            "geqo_effort",
            "force_parallel_mode",
            "enable_seqscan",
            "enable_nestloop",
        ]

        values = get_configuration_values(self.database, parameters)

        self.log.debug("Current database configuration parameters.")
        for parameter in values:
            log_str = f"{parameter} = {values[parameter]}"
            self.log.debug(log_str)

    def update_model_graph(self):
        """Updates the model graph using models as nodes and fks as edges."""

        self.model_graph = networkx.Graph()
        self.model_graph.add_node(self._temp_table, model=TempCatalog)

        all_models = list(self.models.values()) + list(self.extra_nodes.values())

        for model in all_models:
            table_name = model._meta.table_name
            self.model_graph.add_node(table_name, model=model)

        for model in all_models:
            table_name = model._meta.table_name
            for fk_model in model._meta.model_refs:
                ref_table_name = fk_model._meta.table_name
                if ref_table_name not in self.model_graph.nodes:
                    continue

                # Determines the join weight as the average of the
                # weights of the joined nodes.
                if hasattr(model._meta, "xmatch"):
                    model_weight = model._meta.xmatch.join_weight
                else:
                    model_weight = 1

                if hasattr(fk_model._meta, "xmatch"):
                    fk_model_weight = fk_model._meta.xmatch.join_weight
                else:
                    fk_model_weight = 1

                join_weight = 0.5 * (model_weight + fk_model_weight)

                self.model_graph.add_edge(table_name, ref_table_name, join_weight=join_weight)

            if model in self.models.values():
                rel_model = self.get_relational_model(model, sandboxed=False)
                rel_model._meta.schema = self.schema
                rel_model_tname = rel_model._meta.table_name
                self.model_graph.add_node(rel_model_tname, model=rel_model)
                self.model_graph.add_edge(self._temp_table, rel_model_tname)
                self.model_graph.add_edge(table_name, rel_model_tname)

        return self.model_graph

    def set_process_order(self, order="hierarchical", key="row_count", start_node=None):
        """Sets and returns the order in which tables will be processed.

        See `.XMatchPlanner` for details on how the order is decided depending
        on the input parameters.

        """

        if isinstance(order, (list, tuple)):
            self.log.info(f"processing order: {order}")
            self.process_order = order
            return order

        assert order in ["hierarchical", "global"], f"invalid order {order!r}."
        self.log.info(f"processing order mode is {order!r}")

        assert key in ["row_count", "resolution"], f"invalid key {key}."
        self.log.info(f"ordering key is {key!r}.")

        graph = self.model_graph.copy()

        for model in self.extra_nodes.values():
            graph.remove_node(model._meta.table_name)

        if order == "hierarchical":
            subgraphs = networkx.connected_components(graph)
        else:
            subgraphs = [node for node in graph.nodes]

        subgraphs_ext = []
        for sg in subgraphs:
            if start_node and start_node in sg:
                # Last item in record is 0 for initial table, 1 for other.
                # This prioritises the initial table in a sort without
                # having to reverse.
                subgraphs_ext.append((sg, numpy.nan, 0))
            else:
                if key == "row_count":
                    total_row_count = sum([self.models[tn]._meta.xmatch.row_count for tn in sg])
                    # Use -total_row_count to avoid needing reverse order.
                    subgraphs_ext.append((sg, -total_row_count, 1))
                elif key == "resolution":
                    resolution = [self.models[tn]._meta.xmatch.resolution for tn in sg]
                    if all(numpy.isnan(resolution)):
                        min_resolution = numpy.nan
                    else:
                        min_resolution = numpy.nanmin(resolution)
                    subgraphs_ext.append((sg, -(min_resolution or numpy.nan), 1))

        subgraphs_ordered = list(zip(*sorted(subgraphs_ext, key=lambda x: (x[2], x[1], x[0]))))[0]

        if order == "global":
            ordered_tables = subgraphs_ordered
        else:
            ordered_tables = []
            for sgo in subgraphs_ordered:
                sg_ext = []
                for table_name in sgo:
                    model = self.models[table_name]
                    if start_node and start_node == table_name:
                        sg_ext.append((table_name, numpy.nan, 0))
                        continue
                    if key == "row_count":
                        row_count = model._meta.xmatch.row_count
                        sg_ext.append((table_name, -row_count, 1))
                    elif key == "resolution":
                        resolution = model._meta.xmatch.resolution
                        sg_ext.append((table_name, resolution, 1))
                # Use table name as second sorting order
                #  to use alphabetic order in case of draw.
                sg_ordered = list(zip(*sorted(sg_ext, key=lambda x: (x[2], x[1], x[0]))))[0]
                ordered_tables.extend(sg_ordered)

        self.log.info(f"processing order: {ordered_tables}")
        self.process_order = ordered_tables

        return ordered_tables

    def get_join_paths(self, source, return_models=False, mode="full"):
        """Determines all possible join path between two tables.

        Mode ``original``  follows a scorched earth approach
        in which once an edge has been used for a join it cannot be used again.
        This produces only distinct joins between the two nodes.
        Mode ``full`` includes all possible paths that are not a subsample
        of another path.
        In both modes paths that include only the source and destination
        through their relational table are ignored, as are paths in which
        the last node before the output table has not yet been processed.
        Finally mode "config_list" is used to take the paths from a list
        indicated in the configuration file.

        Weights can be defined by setting the ``join_weight`` value in
        `.XMatchModel`. The weight for each edge is the average of the the
        ``join_weight`` of the two nodes joined. The weight defaults to 1.
        Lower weights translate to better chances of that join path to be
        selected.

        Parameters
        ----------
        source : str
            The initial table for the path.
        return_models : bool
            If `True`, returns each path as a list of models. Otherwise returns
            the table names.
        mode : str
            The method used to obtain the paths. Possible values are
            ``original``, ``full``, and ``config_list``

        Returns
        -------
        `list`
            A list in which each item is a path with the table names or models
            joining ``source`` to ``dest``. The list is sorted in order of
            increasing path length.

        """

        graph = self.model_graph.copy()
        porder = self.process_order
        paths = []

        if mode == "original":
            dest = self._temp_table
            rel_table_name = self._temp_table + "_to_" + source
            if rel_table_name in graph:
                graph.remove_node(rel_table_name)

            while True:
                try:
                    spath = shortest_path(graph, source, dest, weight="join_weight")
                    if spath[-3] in porder and porder.index(spath[-3]) < porder.index(source):
                        paths.append(spath)
                    graph.remove_node(spath[1])
                except networkx.NetworkXNoPath:
                    break

        if mode == "full":
            ind = porder.index(source)
            for dest in porder[:ind]:
                paths1 = list(networkx.all_simple_paths(graph, source, dest, cutoff=1))
                if len(paths1) > 0:
                    paths += paths1
                    graph.remove_node(paths1[0][1])  # Only 1 possible path
            for dest in porder[:ind]:
                paths2orless = list(networkx.all_simple_paths(graph, source, dest, cutoff=2))
                paths2 = [el for el in paths2orless if len(el) == 3]
                if len(paths2) > 0:
                    paths += paths2
                    used2_edges = [(el[1], el[2]) for el in paths2]
                    graph.remove_edges_from(used2_edges)
            for dest in porder[:ind]:
                forb_couples = [set(el[1:]) for el in paths if len(el) == 3]
                paths3orless = list(networkx.all_simple_paths(graph, source, dest, cutoff=3))
                paths3 = [
                    el
                    for el in paths3orless
                    if len(el) == 4 and {el[1], el[3]} not in forb_couples
                ]
                if len(paths3) > 0:
                    paths += paths3

            full_paths = []
            for path in paths:
                full_path = path + ["catalog_to_" + path[-1], self._temp_table]
                full_paths.append(full_path)
            paths = full_paths

        if mode == "config_list":
            all_paths = self.join_paths
            paths = [path for path in all_paths if path[0] == source]

        if len(paths) == 0:
            return []

        if return_models:
            nodes = self.model_graph.nodes
            paths = [[nodes[node]["model"] for node in path] for path in paths]

        return paths

    def _prepare_models(self):
        """Prepare the Catalog, CatalogTemp, and Version models."""

        # Sets the metadata of the Catalog, TempCatalog, and Version tables.

        Catalog._meta.schema = self.schema
        Catalog._meta.table_name = self.output_table
        Catalog._meta.set_database(self.database)

        TempCatalog._meta.schema = self.temp_schema
        TempCatalog._meta.table_name = self._temp_table
        TempCatalog._meta.set_database(self.database)

        Version._meta.schema = self.schema
        Version._meta.set_database(self.database)

    def _check_version(self, model, force=False):
        """Checks if a model contains a plan version."""

        with self.database.atomic():
            self.database.execute_sql("SET LOCAL enable_seqscan = off;")
            vexists = (
                peewee.Select(
                    columns=[
                        fn.EXISTS(
                            model.select(SQL("1")).where(model.version_id == self.version_id)
                        )
                    ]
                )
                .tuples()
                .execute(self.database)
            )[0][0]

        if vexists:
            msg = (
                f"{model._meta.table_name!r} contains records for this "
                f"cross-matching plan ({self.plan})."
            )

            if force:
                self.log.warning(msg)
            else:
                raise XMatchError(msg)

    def _create_models(self, force=False):
        """Creates models and performs some checks."""

        self._prepare_models()

        # Bind models
        self.database.bind([Catalog, TempCatalog, Version])

        if not Version.table_exists():
            self.database.create_tables([Version])
            self.log.info(f"Created table {Version._meta.table_name}.")

        if self.version_id is None:
            version, vcreated = Version.get_or_create(plan=self.plan, tag=self.tag)
            self.version_id = version.id
        else:
            vcreated = False

        if vcreated:
            vmsg = "Added version record "
        else:
            vmsg = "Using version record "

        self.log.info(vmsg + f"({self.version_id}, {self.plan}, {self.tag}).")

        # Make sure the output table exists.
        if not Catalog.table_exists():
            # Add Q3C index for Catalog
            Catalog.add_index(
                SQL(
                    f"CREATE INDEX catalog_q3c_idx ON "
                    f"{self.schema}.{self.output_table} "
                    f"(q3c_ang2ipix(ra, dec))"
                )
            )

            self.database.create_tables([Catalog])

            self.log.info(f"Created table {self.output_table}.")

        # Check if Catalog already has entries for this xmatch version.
        if Catalog.table_exists():
            self._check_version(Catalog, force or self.is_addendum)

        if TempCatalog.table_exists():
            msg = f"Temporary table {self._temp_table} already exists."

            if force:
                self.log.warning(msg)
            else:
                raise XMatchError(msg)

            self._check_version(TempCatalog, force)

            try:
                self._temp_count = int(
                    get_row_count(
                        self.database,
                        self._temp_table,
                        schema=self.schema,
                        approximate=True,
                    )
                )
            except ValueError:
                self._temp_count = 0

        else:
            # Add Q3C index for TempCatalog
            TempCatalog.add_index(
                SQL(
                    f"CREATE INDEX IF NOT EXISTS "
                    f"{self._temp_table}_q3c_idx ON "
                    f"{self.temp_schema}.{self._temp_table} "
                    f"(q3c_ang2ipix(ra, dec))"
                )
            )

            self.database.create_tables([TempCatalog])

            self.log.info(f"Created table {self._temp_table}.")

            self._temp_count = 0

    def run(
        self,
        vacuum=False,
        analyze=False,
        from_=None,
        force=False,
        dry_run=False,
        keep_temp=False,
    ):
        """Runs the cross-matching process.

        Parameters
        ----------
        vacuum : bool
            Vacuum all output tables before processing new catalogues.
        analyze : bool
            Analyze all output tables before processing new catalogues.
        from_ : str
            Table from which to start running the process. Useful in reruns
            to skip tables already processed.
        force : bool
            Allows to continue even if the temporary table exists or the
            output table contains records for this version. ``force=True`` is
            assumed if ``from_`` is defined.
        dry_run : bool
            If `False`, loads the temporary tables into ``Catalog`` and ``CatalogToXXX``.
            `True` implies ``keep_temp=True``; all the cross-matching steps will be run
            but the original tables won't be modified. A dry run can only be executed for
            a plan with a single catalogue since processing multiple catalogue requires the
            final tables to have been updated for successive catalogues.
        keep_temp : bool
            Whether to keep the temporary table or to drop it after the cross
            matching is done.

        """

        if len(self.process_order) > 1 and dry_run is True:
            raise RuntimeError("Cannot dry run with a plan that includes more than one catalogue.")

        if dry_run:
            keep_temp = True

        if vacuum or analyze:
            cmd = " ".join(("VACUUM" if vacuum else "", "ANALYZE" if analyze else "")).strip()
            self.log.info(f"Running {cmd} on output tables.")
            vacuum_outputs(
                self.database,
                vacuum=vacuum,
                analyze=analyze,
                schema=Catalog._meta.schema,
                table=Catalog._meta.table_name,
            )
            if TempCatalog.table_exists():
                vacuum_table(self.database, f"{self.schema}.{self._temp_table}")

        # Checks if there are multiple cross-matching plans running at the same
        # time. This is problematic because there can be catalogid collisions.
        # This should only be allowed if we define the starting catalogid
        # manually and make sure there cannot be collisions.
        # temp_tables = [table for table in self.database.get_tables(self.schema)
        #                if table.startswith(self.output_table + '_') and
        #                not table.startswith(self.output_table + '_to_') and
        #                table != self._temp_table]

        # if len(temp_tables) > 0:
        #     raise XMatchError('Another cross-match plan is currently running.')

        self._create_models(force or (from_ is not None))

        if from_ is not None:
            max_cid = TempCatalog.select(fn.MAX(TempCatalog.catalogid)).scalar()
            self._max_cid = max_cid + 1  # just to be sure numbers dont overlap

        with Timer() as timer:
            p_order = self.process_order
            for norder, table_name in enumerate(p_order):
                if dry_run and norder > 0:
                    raise RuntimeError("Cannot dry run more than one catalogue.")

                if from_ and p_order.index(table_name) < p_order.index(from_):
                    self.log.warning(f"Skipping table {table_name}.")
                    continue
                model = self.models[table_name]
                self.process_model(model, force=force)

                if not dry_run:
                    self.load_output_tables(model, keep_temp=keep_temp)

        self.log.info(f"Cross-matching completed in {timer.interval:.3f} s.")

    def process_model(self, model, force=False):
        """Processes a model, loading it into the output table."""

        table_name = model._meta.table_name
        self.log.header = f"[{table_name.upper()}] "

        self.log.info(f"Processing table {table_name}.")
        self._log_table_configuration(model)

        if model._meta.xmatch.has_duplicates:
            raise TargetSelectionNotImplemented(
                "handling of tables with duplicates is not implemented."
            )

        rel_model = self.get_relational_model(model, sandboxed=True, create=False)
        rel_model_table_name = rel_model._meta.table_name
        if rel_model.table_exists():
            if force is False:
                raise RuntimeError(
                    f"Sandboxed relational table {rel_model_table_name} exists. "
                    "Delete it manually before continuing."
                )
            else:
                self.log.warning(f"Sandboxed relational table {rel_model_table_name} exists.")

        # Check if there are already records in catalog for this version.
        if self.process_order.index(model._meta.table_name) == 0 and not self.is_addendum:
            is_first_model = True
        else:
            is_first_model = False

        self._phases_run = set()

        with Timer() as timer:
            if is_first_model:
                self._run_phase_3(model)
            else:
                self._run_phase_1(model)
                self._run_phase_2(model)
                self._run_phase_3(model)

        self.log.info(f"Fully processed {table_name} in {timer.elapsed:.0f} s.")

        self.update_model_graph()

        self.log.header = ""

    def _get_model_fields(self, model):
        """Returns the model fields needed to populate Catalog."""

        meta = model._meta
        xmatch = meta.xmatch
        fields = meta.fields

        # List of fields that will become the SELECT clause.
        model_fields = []

        ra_field = fields[xmatch.ra_column]
        dec_field = fields[xmatch.dec_column]

        to_epoch = self._options["epoch"]

        # RA, Dec, and proper motion fields.
        if model._meta.table_name == "tic_v8":
            # TODO: this should be handled in a way that can be opted-in from
            # the configuration, but for now I'll just hardcode it here.

            pmra_field = fields[xmatch.pmra_column]
            pmdec_field = fields[xmatch.pmdec_column]
            delta_years2000 = to_epoch - get_epoch(model)
            delta_years2015p5 = to_epoch - 2015.5

            racorr2000_field, deccorr2000_field = sql_apply_pm(
                ra_field,
                dec_field,
                pmra_field,
                pmdec_field,
                delta_years2000,
                xmatch.is_pmra_cos,
            )

            racorr2015p5_field, deccorr2015p5_field = sql_apply_pm(
                model.ra_orig,
                model.dec_orig,
                pmra_field,
                pmdec_field,
                delta_years2015p5,
                xmatch.is_pmra_cos,
            )

            ra_field = Case(
                None,
                [(model.posflag == "gaia2", racorr2015p5_field)],
                racorr2000_field,
            )
            dec_field = Case(
                None,
                [(model.posflag == "gaia2", deccorr2015p5_field)],
                deccorr2000_field,
            )

            model_fields.extend([ra_field.alias("ra"), dec_field.alias("dec")])
            model_fields.extend([pmra_field.alias("pmra"), pmdec_field.alias("pmdec")])

        elif xmatch.pmra_column:
            pmra_field = fields[xmatch.pmra_column]
            pmdec_field = fields[xmatch.pmdec_column]

            if (xmatch.epoch and xmatch.epoch != to_epoch) or xmatch.epoch_column:
                delta_years = to_epoch - get_epoch(model)
                ra_field, dec_field = sql_apply_pm(
                    ra_field,
                    dec_field,
                    pmra_field,
                    pmdec_field,
                    delta_years,
                    xmatch.is_pmra_cos,
                )
                if not xmatch.is_pmra_cos:
                    pmra_field *= fn.cos(fn.radians(dec_field))

            model_fields.extend([ra_field.alias("ra"), dec_field.alias("dec")])
            model_fields.extend([pmra_field.alias("pmra"), pmdec_field.alias("pmdec")])

        else:
            pmra_field = peewee.SQL("null")
            pmdec_field = peewee.SQL("null")

            model_fields.extend([ra_field.alias("ra"), dec_field.alias("dec")])

        # Parallax
        if xmatch.parallax_column:
            model_fields.append(fields[xmatch.parallax_column].alias("parallax"))

        return model_fields

    def get_output_model(self, temporary=False):
        """Returns the temporary or final output model (``catalog``)."""

        if temporary:
            return TempCatalog

        return Catalog

    def get_relational_model(self, model, sandboxed=False, temp=False, create=False):
        """Gets or creates a relational table for a given model.

        When the relational model is ``sandboxed``, the table is created in the
        temporary schema and suffixed with the same MD5 used for the run.

        """

        cat_table = Catalog._meta.table_name
        prefix = cat_table + "_to_"

        meta = model._meta
        pk = meta.primary_key

        if isinstance(pk, str) and pk == "__composite_key__":
            raise XMatchError(f"composite pk found for model {model.__name__!r}.")

        # Auto/Serial are automatically PKs. Convert them to integers
        # to avoid having two pks in the relational table.
        if pk.__class__.field_type == "AUTO":
            model_pk_class = peewee.IntegerField
        elif pk.__class__.field_type == "BIGAUTO":
            model_pk_class = peewee.BigIntegerField
        else:
            model_pk_class = pk.__class__

        class BaseModel(peewee.Model):
            catalogid = peewee.BigIntegerField(null=False, index=True)
            target_id = model_pk_class(null=False, index=True)
            version_id = peewee.SmallIntegerField(null=False, index=True)
            distance = peewee.DoubleField(null=True)
            best = peewee.BooleanField(null=False)
            plan_id = peewee.TextField(null=True)
            added_by_phase = peewee.SmallIntegerField(null=True)

            class Meta:
                database = meta.database
                schema = self.temp_schema if sandboxed else meta.schema
                primary_key = False

        model_prefix = "".join(x.capitalize() or "_" for x in prefix.rstrip().split("_"))

        RelationalModel = type(model_prefix + model.__name__, (BaseModel,), {})

        if temp:
            RelationalModel._meta.primary_key = False
            RelationalModel._meta.composite_key = False
            RelationalModel._meta.set_table_name(uuid.uuid4().hex[0:8])
            RelationalModel._meta.schema = None
            return RelationalModel

        table_name = prefix + meta.table_name
        if sandboxed:
            table_name += f"_{self.md5}"
        RelationalModel._meta.table_name = table_name

        if create and not RelationalModel.table_exists():
            RelationalModel.create_table()

        # Add foreign key field here. We want to avoid Peewee creating it
        # as a constraint and index if the table is created because that would
        # slow down inserts. We'll created them manually with add_fks.
        # Note that we do not create an FK between the relational model and
        # Catalog because the relationship is only unique on
        # (catalogid, version_id).
        RelationalModel._meta.remove_field("target_id")
        RelationalModel._meta.add_field(
            "target_id",
            peewee.ForeignKeyField(model, column_name="target_id", backref="+"),
        )

        return RelationalModel

    def _build_join(self, path):
        """Returns a build query for a given join path."""

        model = path[0]

        query = model.select()
        for inode in range(1, len(path)):
            if path[inode] is TempCatalog:
                query = query.join(
                    TempCatalog, on=(TempCatalog.catalogid == path[inode - 1].catalogid)
                )
            elif path[inode]._meta.table_name == "gaia_dr2_neighbourhood":
                query = (
                    query.join(path[inode])
                    .where(path[inode].angular_distance < 200)  # For 1-1
                    .distinct(path[inode].dr2_source_id)
                )  # To confirm
            else:
                query = query.join(path[inode])

        return query

    def _run_phase_1(self, model):
        """Runs the linking against matched catalogids stage."""

        table_name = model._meta.table_name
        rel_model_sb = self.get_relational_model(model, create=True, sandboxed=True)

        rel_model = self.get_relational_model(model, create=False, sandboxed=False)

        model_pk = model._meta.primary_key

        self.log.info("Phase 1: linking existing targets.")

        if 1 in model._meta.xmatch.skip_phases:
            self.log.warning("Skipping due to configuration.")
            return
        path_mode = self.path_mode
        join_paths = self.get_join_paths(table_name, mode=path_mode)

        if len(join_paths) == 0:
            self.log.debug(f"No paths found between {table_name} and temporary output table.")
            return False

        self.log.debug(
            f"Found {len(join_paths)} paths between {table_name} and temporary output table."
        )

        for n_path, path in enumerate(join_paths):
            # Remove the temporary catalog table at the end of the join path
            # because we only need catalogid and we can get that from the
            # relational model, saving us one join.
            path = path[0:-1]

            join_models = [self.model_graph.nodes[node]["model"] for node in path]

            # Get the relational model that leads to the temporary catalog
            # table in the join. We'll want to filter on version_id to avoid
            # a sequential scan.
            join_rel_model = join_models[-1]

            query = (
                self._build_join(join_models)
                .select(
                    model_pk.alias("target_id"),
                    join_rel_model.catalogid,
                    peewee.Value(True).alias("best"),
                )
                .where(
                    join_rel_model.version_id == self.version_id,
                    join_rel_model.best >> True,
                )
                .where(
                    ~fn.EXISTS(
                        rel_model_sb.select(SQL("1")).where(
                            rel_model_sb.version_id == self.version_id,
                            (
                                (rel_model_sb.target_id == model_pk)
                                | (rel_model_sb.catalogid == join_rel_model.catalogid)
                            ),
                        )
                    )
                )
                # Select only one match per target in the catalogue with are cross-matching.
                .distinct(model_pk)
            )

            # Deal with duplicates in LS8
            if table_name == "legacy_survey_dr8":
                query = query.where(self._get_ls8_where(model))

            # Remove Duplicates and non-primary entries from LS10
            if table_name == "legacy_survey_dr10":
                query = query.where(
                    model.survey_primary >> True, fn.coalesce(model.ref_cat, "") != "T2"
                )

            # If the real relational model exists, exclude any matches that already exist there.
            if rel_model.table_exists():
                query = query.where(
                    ~fn.EXISTS(
                        rel_model.select(SQL("1")).where(
                            rel_model.version_id == self.version_id,
                            rel_model.target_id == model_pk,
                        )
                    )
                )

            # In query we do not include a Q3C where for the sample region because
            # TempCatalog for this plan should already be sample region limited.

            with Timer() as timer:
                with self.database.atomic():
                    temp_model = self.get_relational_model(model, temp=True, sandboxed=True)
                    temp_table = temp_model._meta.table_name

                    self._setup_transaction(model, phase=1)
                    self.log.debug(
                        f"Selecting linked targets into temporary "
                        f"table {temp_table!r} with join path "
                        f"{path}{self._get_sql(query)}"
                    )
                    query.create_table(temp_table, temporary=True)

                    self.log.debug(
                        f"Copying data into relational model "
                        f"{rel_model_sb._meta.table_name!r}."
                    )

                    fields = [
                        temp_model.target_id,
                        temp_model.catalogid,
                        temp_model.version_id,
                        temp_model.best,
                        temp_model.plan_id,
                        temp_model.added_by_phase,
                    ]

                    nids = (
                        rel_model_sb.insert_from(
                            temp_model.select(
                                temp_model.target_id,
                                temp_model.catalogid,
                                peewee.Value(self.version_id),
                                temp_model.best,
                                peewee.Value(self.plan) if self.is_addendum else None,
                                peewee.Value(1),
                            ),
                            fields,
                        )
                        .returning()
                        .execute()
                    )

            self.log.debug(f"Linked {nids.rowcount:,} records in {timer.interval:.3f} s.")
            self._phases_run.add(1)

            if nids.rowcount > 0:
                self._analyze(rel_model_sb)

    def _run_phase_2(self, model, source=TempCatalog):
        """Associates existing targets in Catalog with entries in the model.

        Here ``source`` is the catalogue with which we are spatially cross-matching.
        Normally this is the temporary catalog table which we are building for this
        cross-match run. But when we are doing an addendum, that table is going to be
        empty (at least of the first table in the addeundum), so we need to also have
        the option of using the real ``catalog`` table as the source. This method will
        call itself recursively with ``source=Catalog`` if this is an addendum run.

        """

        meta = model._meta
        xmatch = meta.xmatch

        table_name = meta.table_name

        s_table_name = source._meta.table_name
        self.log.info(f"Phase 2: cross-matching against existing targets ({s_table_name}).")

        if 2 in xmatch.skip_phases:
            self.log.warning("Skipping due to configuration.")
            return

        rel_model_sb = self.get_relational_model(model, create=True, sandboxed=True)
        rel_sb_table_name = rel_model_sb._meta.table_name

        rel_model = self.get_relational_model(model, create=False, sandboxed=False)

        model_pk = meta.primary_key
        model_ra = meta.fields[xmatch.ra_column]
        model_dec = meta.fields[xmatch.dec_column]

        catalog_epoch = self._options["epoch"]

        query_radius = xmatch.query_radius or self._options["query_radius"]

        # Should we use proper motions?
        model_epoch = get_epoch(model)

        is_model_expression = isinstance(model_epoch, (peewee.Expression, peewee.Function))
        use_pm = model_epoch and (is_model_expression or (model_epoch != catalog_epoch))

        if use_pm:
            self.log.debug("Determining maximum epoch delta " "between catalogues.")

            if isinstance(model_epoch, (int, float)):
                max_delta_epoch = float(abs(model_epoch - catalog_epoch))
            else:
                max_delta_epoch = float(
                    model.select(fn.MAX(fn.ABS(model_epoch - catalog_epoch)))
                    .where(self._get_sample_where(model_ra, model_dec))
                    .scalar()
                )

            max_delta_epoch += 0.1  # Add .1 yr to be sure it's an upper bound

            self.log.debug(f"Maximum epoch delta: {max_delta_epoch:.3f} (+ 0.1 year).")

        if use_pm:
            model_pmra = meta.fields[xmatch.pmra_column]
            model_pmdec = meta.fields[xmatch.pmdec_column]
            model_is_pmra_cos = int(xmatch.is_pmra_cos)

            q3c_dist = fn.q3c_dist_pm(
                model_ra,
                model_dec,
                model_pmra,
                model_pmdec,
                model_is_pmra_cos,
                model_epoch,
                source.ra,
                source.dec,
                catalog_epoch,
            )
            q3c_join = fn.q3c_join_pm(
                model_ra,
                model_dec,
                model_pmra,
                model_pmdec,
                model_is_pmra_cos,
                model_epoch,
                source.ra,
                source.dec,
                catalog_epoch,
                max_delta_epoch,
                query_radius / 3600.0,
            )
        else:
            q3c_dist = fn.q3c_dist(model_ra, model_dec, source.ra, source.dec)
            q3c_join = fn.q3c_join(
                model_ra,
                model_dec,
                source.ra,
                source.dec,
                query_radius / 3600.0,
            )

        # Get the cross-matched catalogid and model target pk (target_id),
        # and their distance.
        xmatched = (
            source.select(
                source.catalogid,
                model_pk.alias("target_id"),
                q3c_dist.alias("distance"),
                source.version_id,
            )
            .join(model, peewee.JOIN.CROSS)
            .where(q3c_join)
            .where(self._get_sample_where(model_ra, model_dec))
        )

        if table_name == "legacy_survey_dr8":
            xmatched = xmatched.where(self._get_ls8_where(model))
        if table_name == "legacy_survey_dr10":
            xmatched = xmatched.where(
                model.survey_primary >> True,
                fn.coalesce(model.ref_cat, "") != "T2",
            )

        # This may break the use of the index but I think it's needed if
        # the model is the second table in q3c_join and has empty RA/Dec.
        if xmatch.has_missing_coordinates and use_pm:
            xmatched = xmatched.where(model_ra.is_null(False), model_dec.is_null(False))

        xmatched = xmatched.cte("xmatched", materialized=True)

        # We'll partition over each group of targets that match the same
        # catalogid and mark the one with the smallest distance to it as best.
        partition = fn.first_value(xmatched.c.target_id).over(
            partition_by=[xmatched.c.catalogid],
            order_by=[xmatched.c.distance.asc()],
        )
        best = peewee.Value(partition == xmatched.c.target_id)

        # Select the values to insert. Remove target_ids that were already
        # present in the relational table due to phase 1.
        # We separate the filter in two IF NOT EXISTS clauses to
        # be sure the query planner uses the indexes for each (it
        # won't necessarily do it if we do an OR). Also, make sure
        # we compare version_id, target_id and not the other way
        # because that's the order in which we defined the index.
        in_query = xmatched.select(
            xmatched.c.target_id,
            xmatched.c.catalogid,
            peewee.Value(self.version_id).alias("version_id"),
            xmatched.c.distance.alias("distance"),
            best.alias("best"),
            self.plan if self.is_addendum else None,
            peewee.Value(2).alias("added_by_phase"),
        )

        # We only need to care about already linked targets if phase 1 run.
        if 1 in self._phases_run:
            in_query = in_query.where(
                xmatched.c.version_id == self.version_id,
                ~fn.EXISTS(
                    rel_model_sb.select(SQL("1")).where(
                        (rel_model_sb.version_id == self.version_id)
                        & (
                            (rel_model_sb.catalogid == xmatched.c.catalogid)
                            | (rel_model_sb.target_id == xmatched.c.target_id)
                        )
                    )
                ),
            )

        if rel_model.table_exists():
            in_query = in_query.where(
                ~fn.EXISTS(
                    rel_model.select(SQL("1")).where(
                        (rel_model.version_id == self.version_id)
                        & (
                            (rel_model.catalogid == xmatched.c.catalogid)
                            | (rel_model.target_id == xmatched.c.target_id)
                        )
                    )
                )
            )

        # Make sure we are only spatially cross-matching against targets from the same
        # cross-match. This only matters for addendum runs because in those we are also
        # cross-matching against the existing catalogdb.catalog table which contains multiple
        # versions.
        in_query = in_query.where(xmatched.c.version_id == self.version_id)

        with Timer() as timer:
            with self.database.atomic():
                # 1. Tweak database configuration for this transaction to
                #    ensure Q3C index is used.

                # May need to increase work_mem during this transaction to
                # make sure the Q3C index is used.
                self._setup_transaction(model, phase=2)

                # 2. Run cross-match and insert data into relational model.

                fields = [
                    rel_model_sb.target_id,
                    rel_model_sb.catalogid,
                    rel_model_sb.version_id,
                    rel_model_sb.distance,
                    rel_model_sb.best,
                    rel_model_sb.plan_id,
                    rel_model_sb.added_by_phase,
                ]

                in_query = rel_model_sb.insert_from(
                    in_query.with_cte(xmatched),
                    fields,
                ).returning()

                self.log.debug(
                    f"Running Q3C query and inserting cross-matched data into "
                    f"relational table {rel_sb_table_name!r}: "
                    f"{self._get_sql(in_query)}"
                )

                n_catalogid = in_query.execute().rowcount

        self.log.debug(
            f"Cross-matched {source._meta.table_name} with "
            f"{n_catalogid:,} targets in {table_name}. "
            f"Run in {timer.interval:.3f} s."
        )

        if n_catalogid > 0:
            self._phases_run.add(2)
            self._analyze(rel_model_sb)

        # For addenda it's not sufficient to cross-match with the temporary table, because that
        # does not contain all the cumulated targets from this cross-match version. We need to
        # also spatially cross-match with the real catalog table (but only for the targets with
        # version_id=<this-version-id>).
        if self.is_addendum and source != Catalog:
            self._run_phase_2(model, source=Catalog)

    def _run_phase_3(self, model):
        """Add non-matched targets to Catalog and the relational table."""

        meta = model._meta
        xmatch = meta.xmatch

        self.log.info("Phase 3: adding non cross-matched targets.")

        rel_model_sb = self.get_relational_model(model, create=True, sandboxed=True)
        rel_sb_table_name = rel_model_sb._meta.table_name

        rel_model = self.get_relational_model(model, create=False, sandboxed=False)

        if 3 in xmatch.skip_phases:
            self.log.warning("Skipping due to configuration.")
            return

        table_name = meta.table_name

        model_fields = self._get_model_fields(model)

        model_pk = meta.primary_key
        model_ra = meta.fields[xmatch.ra_column]
        model_dec = meta.fields[xmatch.dec_column]

        unmatched = model.select(
            (fn.row_number().over() + self._max_cid).alias("catalogid"),
            model_pk.alias("target_id"),
            *model_fields,
        ).where(self._get_sample_where(model_ra, model_dec))

        if 1 in self._phases_run or 2 in self._phases_run:
            unmatched = unmatched.where(
                ~fn.EXISTS(
                    rel_model_sb.select(SQL("1")).where(
                        rel_model_sb.version_id == self.version_id,
                        rel_model_sb.target_id == model_pk,
                        rel_model_sb.best >> True,
                    )
                )
            )

        if rel_model.table_exists():
            unmatched = unmatched.where(
                ~fn.EXISTS(
                    rel_model.select(SQL("1")).where(
                        rel_model.version_id == self.version_id,
                        rel_model.target_id == model_pk,
                        rel_model.best >> True,
                    )
                )
            )

        if xmatch.has_missing_coordinates:
            unmatched = unmatched.where(model_ra.is_null(False), model_dec.is_null(False))

        # TODO: this is horrible and should be moved to the configuration.
        if model._meta.table_name == "tic_v8":
            unmatched = unmatched.where(model.objtype != "EXTENDED")

        if table_name == "legacy_survey_dr8":
            unmatched = unmatched.where(self._get_ls8_where(model))

        if table_name == "legacy_survey_dr10a":
            unmatched = unmatched.where(model.ref_cat != "T2")

        if table_name == "legacy_survey_dr10":
            unmatched = unmatched.where(
                model.survey_primary >> True, fn.coalesce(model.ref_cat, "") != "T2"
            )

        with Timer() as timer:
            with self.database.atomic():
                # TODO: Not sure it's worth using a temporary table here.

                # 1. Run link query and create temporary table with results.

                self._setup_transaction(model, phase=3)

                temp_model = self.get_relational_model(model, temp=True, sandboxed=True)
                temp_model_name = temp_model._meta.table_name

                self.log.debug(
                    f"Selecting unique targets "
                    f"into temporary table "
                    f"{temp_model_name!r}{self._get_sql(unmatched)}"
                )

                unmatched.create_table(temp_model_name, temporary=True)

                # Analyze the temporary table to gather stats.
                # self.log.debug('Running ANALYZE on temporary table.')
                # self.database.execute_sql(f'ANALYZE "{temp_model_name}";')

                # 2. Copy data from temporary table to relational table. Add
                #    catalogid at this point.

                fields = [
                    temp_model.catalogid,
                    temp_model.target_id,
                    temp_model.version_id,
                    temp_model.best,
                    rel_model_sb.plan_id,
                    rel_model_sb.added_by_phase,
                ]

                rel_insert_query = rel_model_sb.insert_from(
                    temp_model.select(
                        temp_model.catalogid,
                        temp_model.target_id,
                        self.version_id,
                        peewee.SQL("true"),
                        self.plan if self.is_addendum else None,
                        peewee.Value(3).alias("added_by_phase"),
                    ),
                    fields,
                ).returning()

                self.log.debug(
                    f"Copying data into relational model "
                    f"{rel_sb_table_name!r}"
                    f"{self._get_sql(rel_insert_query)}"
                )

                cursor = rel_insert_query.execute()
                n_rows = cursor.rowcount

                self.log.debug(
                    f"Insertion into {rel_sb_table_name} completed "
                    f"with {n_rows:,} rows in "
                    f"{timer.elapsed:.3f} s."
                )

                # 3. Fill out the temporary catalog table with the information
                #    from the unique targets.

                temp_table = peewee.Table(temp_model_name)

                fields = [
                    TempCatalog.catalogid,
                    TempCatalog.lead,
                    TempCatalog.version_id,
                ]
                select_columns = [
                    temp_table.c.catalogid,
                    peewee.Value(table_name),
                    self.version_id,
                ]
                for field in model_fields:
                    if field._alias == "ra":
                        fields.append(TempCatalog.ra)
                        select_columns.append(temp_table.c.ra)
                    elif field._alias == "dec":
                        fields.append(TempCatalog.dec)
                        select_columns.append(temp_table.c.dec)
                    elif field._alias == "pmra":
                        fields.append(TempCatalog.pmra)
                        select_columns.append(temp_table.c.pmra)
                    elif field._alias == "pmdec":
                        fields.append(TempCatalog.pmdec)
                        select_columns.append(temp_table.c.pmdec)
                    elif field._alias == "parallax":
                        fields.append(TempCatalog.parallax)
                        select_columns.append(temp_table.c.parallax)

                insert_query = TempCatalog.insert_from(
                    temp_table.select(*select_columns), fields
                ).returning()

                self.log.debug(
                    f"Running INSERT query into {self._temp_table}"
                    f"{self._get_sql(insert_query)}"
                )

                cursor = insert_query.execute()
                n_rows = cursor.rowcount

        # Avoid having to calculate max_cid again
        self._max_cid += n_rows
        self._temp_count += n_rows

        self.log.debug(f"Inserted {n_rows:,} rows. Total time: {timer.elapsed:.3f} s.")

        self._phases_run.add(3)

        if n_rows > 0.5 * self._temp_count:  # Cluster if > 50% of rows are new
            self.log.debug(f"Running CLUSTER on {self._temp_table} with q3c index.")
            self.database.execute_sql(
                f"CLUSTER {self.temp_schema}.{self._temp_table} "
                f"using {self._temp_table}_q3c_idx;"
            )
            self.log.debug(f"Running ANALYZE on {self._temp_table}.")
            self.database.execute_sql(f"ANALYZE {self.temp_schema}.{self._temp_table};")

        self._analyze(rel_model_sb, catalog=False)

    def load_output_tables(self, model, keep_temp=False):
        """Loads the temporary tables into the output tables."""

        self._load_output_table(TempCatalog, Catalog, keep_temp=keep_temp)

        rel_model_sb = self.get_relational_model(model, sandboxed=True, create=False)
        rel_model = self.get_relational_model(model, sandboxed=False, create=True)
        self._load_output_table(rel_model_sb, rel_model, keep_temp=keep_temp)

    def _load_output_table(self, from_model, to_model, keep_temp=False):
        """Copies the temporary table to the real output table."""

        to_table = f"{to_model._meta.schema}.{to_model._meta.table_name}"
        from_table = f"{from_model._meta.schema}.{from_model._meta.table_name}"

        self.log.info(f"Copying {from_table} to {to_table}.")

        with Timer() as timer:
            with self.database.atomic():
                self._setup_transaction()

                insert_query = to_model.insert_from(
                    from_model.select(), from_model.select()._returning
                ).returning()

                self.log.debug(
                    f"Running INSERT query into {to_table} {self._get_sql(insert_query)}"
                )

                cursor = insert_query.execute()
                n_rows = cursor.rowcount

        self.log.debug(f"Inserted {n_rows:,} rows in {timer.elapsed:.3f} s.")

        if not keep_temp:
            self.database.drop_tables([from_model])
            self.log.info(f"Dropped temporary table {from_table}.")

        if n_rows > 0:
            self.log.debug(f"Running VACUUM ANALYZE on {to_table}.")
            vacuum_table(self.database, to_table, vacuum=True, analyze=True)

    def _get_sql(self, query, return_string=False):
        """Returns colourised SQL text for logging."""

        query_str, query_params = query.sql()

        if query_params:
            for ind in range(len(query_params)):
                if isinstance(query_params[ind], str):
                    query_params[ind] = "'" + query_params[ind] + "'"

            query_str = query_str % tuple(query_params)

        query_str = query_str.replace("None", "Null")
        if return_string:
            return query_str
        elif self.log.rich_console:
            return f": {rich.markup.escape(query_str)}"
        elif self._options["show_sql"]:
            return f': {color_text(query_str, "blue")}'
        else:
            return "."

    def _setup_transaction(self, model=None, phase=None):
        """Sets database parameters for the transaction."""

        if not self._options["database_options"]:
            return

        options = self._options["database_options"].copy()
        if model and model._meta.xmatch.database_options:
            options.update(model._meta.xmatch.database_options)

        for param in options:
            if param == "maintenance_work_mem":
                continue
            param_config = options[param]
            if isinstance(param_config, dict):
                value = param_config.get("value")
                param_phases = param_config.get("phase", None)
                if param_phases and (not phase or phase not in param_phases):
                    continue
            else:
                value = param_config
            stm = f"SET LOCAL {param}={value!r};"
            self.database.execute_sql(stm)
            self.log.debug(stm)

    def _get_sample_where(self, ra_field, dec_field):
        """Returns the list of conditions to sample a model."""

        sample_region = self._options["sample_region"]

        if sample_region is None:
            return True

        if len(sample_region) == 3 and not isinstance(sample_region[0], (list, tuple)):
            return fn.q3c_radial_query(
                ra_field,
                dec_field,
                sample_region[0],
                sample_region[1],
                sample_region[2],
            )

        sample_conds = peewee.SQL("false")

        for ra, dec, radius in sample_region:
            sample_conds |= fn.q3c_radial_query(ra_field, dec_field, ra, dec, radius)

        return sample_conds

    def _get_ls8_where(self, model):
        """Removes duplicates from LS8 queries."""

        return ~(
            (
                (model.release == 8000)
                & (model.dec > 32.375)
                & (model.ra > 100.0)
                & (model.ra < 300.0)
            )
            | ((model.release == 8001) & (model.dec < 32.375))
        )

    def show_join_paths(self):
        """Prints all the available joint paths.

        This is useful before call `.run` to make sure the join paths to be
        used are correct or adjust the table weights otherwise. Note that
        the paths starting from the first model to be processed are ignored.

        """
        if len(self.process_order) == 1:
            return

        mode = self.path_mode
        for table in self.process_order[1:]:
            paths = self.get_join_paths(table, mode=mode)
            if paths:
                for path in paths:
                    print(path)

    def _analyze(self, rel_model, vacuum=False, catalog=False):
        """Analyses a relational model after insertion."""

        schema = rel_model._meta.schema
        table_name = rel_model._meta.table_name

        db_opts = self._options["database_options"]
        if db_opts:
            work_mem = db_opts.get("maintenance_work_mem", None)
            if work_mem:
                self.database.execute_sql(f"SET maintenance_work_mem = {work_mem!r}")

        self.log.debug(f"Running ANALYZE on {table_name}.")
        vacuum_table(self.database, f"{schema}.{table_name}", vacuum=vacuum, analyze=True)

        if catalog:
            self.log.debug(f"Running ANALYZE on {self._temp_table}.")
            vacuum_table(
                self.database,
                f"catalogdb.{self._temp_table}",
                vacuum=vacuum,
                analyze=True,
            )

    def _log_table_configuration(self, model):
        """Logs the configuration used to cross-match a table."""

        xmatch = model._meta.xmatch

        parameters = [
            "ra_column",
            "dec_column",
            "pmra_column",
            "pmdec_column",
            "is_pmra_cos",
            "parallax_column",
            "epoch",
            "epoch_column",
            "epoch_format",
            "has_duplicates",
            "skip",
            "skip_phases",
            "query_radius",
            "row_count",
            "resolution",
            "join_weight",
            "has_missing_coordinates",
        ]

        self.log.debug("Table cross-matching parameters:")

        for parameter in parameters:
            value = getattr(xmatch, parameter, None)

            if parameter == "query_radius":
                value = value or QUERY_RADIUS

            if value is True or value is False or value is None:
                value = str(value).lower()

            if parameter == "row_count":
                self.log.debug(f"{parameter}: {value:,}")
            else:
                self.log.debug(f"{parameter}: {value}")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2023-04-28
# @Filename: mwm_magcloud_rgb.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

import dustmaps.sfd
import numpy
import pandas
from astropy.coordinates import SkyCoord
from astropy.units import mas, yr
from dustmaps.sfd import SFDQuery
from gala.coordinates import MagellanicStreamNidever08
from peewee import fn

from sdssdb.peewee.sdss5db.catalogdb import CatalogToGaia_DR3, Gaia_DR3

from target_selection import log
from target_selection.cartons import BaseCarton
from target_selection.cartons.mwm_magcloud_agb import roi_cut
from target_selection.cartons.tools import create_table_as


class MWM_MagCloud_RGB_BOSS(BaseCarton):
    """MWM Magellanic clouds RGBs.

    Definition:

    Select AGB targets in the Magellanic Clouds using Gaia DR3 photometry and astrometry.
    Selection based on color cuts, proper motion cuts and parallax cuts. See wiki
    for David Nidever's code.

    """

    name = "mwm_magcloud_rgb_boss"
    mapper = "MWM"
    category = "science"
    program = "mwm_magcloud"
    instrument = "BOSS"
    priority = 2819
    cadence = "bright_1x1"
    can_offset = True

    def build_query(self, version_id, query_region=None):
        # Parallax cut
        parallax_cut = ~(
            (Gaia_DR3.parallax > 0) & ((Gaia_DR3.parallax + 0.025) / Gaia_DR3.parallax_error > 5)
        )

        # Rough cuts. Just to reduce the number of rows returned for post-process.
        bp_rp = Gaia_DR3.phot_bp_mean_mag - Gaia_DR3.phot_rp_mean_mag
        colour_cut = (Gaia_DR3.phot_g_mean_mag <= self.parameters["gaiag_rough"]) & (
            bp_rp >= self.parameters["bp_rp_rough"]
        )

        pm_cut = (
            (Gaia_DR3.pmra >= self.parameters["pmra_rough"][0])
            & (Gaia_DR3.pmra <= self.parameters["pmra_rough"][1])
            & (Gaia_DR3.pmdec >= self.parameters["pmdec_rough"][0])
            & (Gaia_DR3.pmdec <= self.parameters["pmdec_rough"][1])
        )

        ra, dec, radius = self.parameters["astro_rough"]
        astro_cut = fn.q3c_radial_query(Gaia_DR3.ra, Gaia_DR3.dec, ra, dec, radius)

        log.info("Running Q3C query ...")
        tmp_q3c, _ = create_table_as(
            Gaia_DR3.select(Gaia_DR3.source_id).where(astro_cut),
            "tmp_q3c",
            temporary=True,
            database=self.database,
        )

        query = (
            CatalogToGaia_DR3.select(
                CatalogToGaia_DR3.catalogid,
                Gaia_DR3.ra,
                Gaia_DR3.dec,
                Gaia_DR3.l,
                Gaia_DR3.b,
                Gaia_DR3.pmra,
                Gaia_DR3.pmdec,
                Gaia_DR3.parallax,
                Gaia_DR3.parallax_error,
                Gaia_DR3.phot_g_mean_mag,
                Gaia_DR3.phot_bp_mean_mag,
                Gaia_DR3.phot_rp_mean_mag,
            )
            .join(Gaia_DR3)
            .join(tmp_q3c, on=(tmp_q3c.c.source_id == Gaia_DR3.source_id))
            .where(
                CatalogToGaia_DR3.version_id == version_id,
                CatalogToGaia_DR3.best >> True,
                Gaia_DR3.phot_g_mean_mag < self.parameters["g_lim"],
                parallax_cut,
                colour_cut,
                pm_cut,
            )
        )

        return query

    def post_process(self, model, **kwargs):
        """Runs post-process."""

        data = pandas.read_sql(f"SELECT * from {self.path}", self.database)

        # Calculate Magellanic Stream coordinates
        coords = SkyCoord(
            ra=data.ra.values,
            dec=data.dec.values,
            unit="deg",
            frame="icrs",
            pm_ra_cosdec=data.pmra.values * mas / yr,
            pm_dec=data.pmdec.values * mas / yr,
        )
        mcoo = coords.transform_to(MagellanicStreamNidever08)

        data["pml_ms"] = mcoo.pm_L_cosB.value
        data["pmb_ms"] = mcoo.pm_B.value
        data["mlat"] = mcoo.B.value
        data["mlon"] = mcoo.L.value

        # Calculate LMC and SMC radius
        lmcrad = numpy.sqrt((data["mlon"] + 0.6) ** 2 + (data["mlat"] - 3.6) ** 2)
        smcrad = numpy.sqrt((data["mlon"] + 15.53) ** 2 + (data["mlat"] + 11.58) ** 2)
        data["lmcrad"] = lmcrad
        data["smcrad"] = smcrad

        # Proper motion cut
        pmdist = numpy.sqrt((data["pml_ms"] - 1.8) ** 2 + (data["pmb_ms"] - 0.40) ** 2)
        gdpm = pmdist < self.parameters["pmdist"]
        data = data.loc[gdpm]

        # Make sure none of the needed quantities are NaN
        good = (
            numpy.isfinite(data["pmra"])
            & numpy.isfinite(data["pmdec"])
            & numpy.isfinite(data["parallax"])
            & numpy.isfinite(data["phot_g_mean_mag"])
            & numpy.isfinite(data["phot_bp_mean_mag"])
            & numpy.isfinite(data["phot_rp_mean_mag"])
        )
        data = data[good]

        # Get extinction values from SFD
        dustmaps.sfd.fetch()
        coo = SkyCoord(ra=data["ra"], dec=data["dec"], unit="degree", frame="icrs")
        sfd = SFDQuery()
        ebv = sfd(coo)

        # For high-extinction regions in the inner LMC/SMC SFD is not that reliable
        # but we can "rescale" ebv to get something reasonable
        # Selection for inner LMC/SMC with high extinction
        (bd,) = numpy.where(((data["lmcrad"] <= 4.5) | (data["smcrad"] <= 2.5)) & (ebv >= 0.1))
        ebv[bd] = ebv[bd] * 0.07  # rescale

        # Get dereddened magnitudes
        Av = 0.86 * 3.1 * ebv
        ag = 0.8509139481578785 * Av
        abp = 1.0901361069987485 * Av
        arp = 0.5894794093231802 * Av
        gmag0 = data["phot_g_mean_mag"] - ag
        bp0 = data["phot_bp_mean_mag"] - abp
        rp0 = data["phot_rp_mean_mag"] - arp

        # Apply CMD cut to dereddened magnitudes
        bprpcut = self.parameters["bprpcut"]
        gcut = self.parameters["gcut"]
        _, cutind = roi_cut(bprpcut, gcut, bp0 - rp0, gmag0)
        data = data.iloc[cutind]

        # Final magnitude cut
        (gcut,) = numpy.where(data["phot_g_mean_mag"] <= 17.5)
        data = data.iloc[gcut]

        valid_cids = data.catalogid.values

        # This way seems faster than updating from a list of values.
        values_cids = ",".join(f"({vc})" for vc in valid_cids)
        self.database.execute_sql(
            "CREATE TEMP TABLE valid_cids AS SELECT * "
            f"FROM (VALUES {values_cids}) "
            "AS t (catalogid)"
        )
        self.database.execute_sql("CREATE INDEX ON valid_cids (catalogid);")
        self.database.execute_sql(f"UPDATE {self.path} SET selected = false")
        self.database.execute_sql(
            f"UPDATE {self.path} SET selected = true "
            "FROM valid_cids vc "
            f"WHERE {self.path}.catalogid = vc.catalogid"
        )

        return super().post_process(model, **kwargs)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: Tom Dwelly
# @Date: 2020-03-03
# @Filename: bhm_aqmes.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

# isort: skip_file

import importlib
import peewee

# from peewee import JOIN
# from peewee import fn
from astropy.io import fits


from sdssdb.peewee.sdss5db.catalogdb import (
    Catalog,
    SDSS_DR19p_Speclite,
    SDSS_DR16_QSO,
    CatalogFromSDSS_DR19p_Speclite,
)

# DEBUG STUFF TO USE TEMP TABLE
# CatalogToSDSS_DR19p_Speclite._meta.table_name = 'temp_catalog_to_sdss_dr19p_speclite'
# CatalogToSDSS_DR19p_Speclite._meta._schema = 'sandbox'

from target_selection.cartons.base import BaseCarton

# this should probably live in a better place
radius_apo = 1.49  # degrees

# Details: Start here
# https://wiki.sdss.org/display/OPS/Defining+target+selection+and+cadence+algorithms#Definingtargetselectionandcadencealgorithms-AQMES-medium-inprogress   # noqa: E501

# This provides the following BHM cartons:
#  bhm_aqmes_med
#  bhm_aqmes_med_faint
#  bhm_aqmes_wide2
#  bhm_aqmes_wide2_faint
#  bhm_aqmes_wide1
#  bhm_aqmes_wide1_faint
#  # bhm_aqmes_wide3        # dumped in v0.5
#  # bhm_aqmes_wide3-faint  # dumped in v0.5
#  bhm_aqmes_bonus_dark
#  bhm_aqmes_bonus_bright

# how do we relate the cadence names in v0.5, 1.0 to cadence names in v0?

cadence_map_v0p5_to_v0 = {
    "dark_10x4": "bhm_aqmes_medium_10x4",
    "dark_10x4_4yr": "bhm_aqmes_medium_10x4",
    # 'dark_3x4': 'bhm_aqmes_wide_3x4',
    "dark_2x4": "bhm_aqmes_wide_2x4",
    "dark_1x4": "bhm_spiders_1x4",
    "bright_3x1": "bhm_boss_bright_3x1",
}
cadence_map_v1_to_v0 = {
    "dark_10x4": "bhm_aqmes_medium_10x4",
    "dark_10x4_4yr": "bhm_aqmes_medium_10x4",
    "dark_2x4": "bhm_aqmes_wide_2x4",
    "dark_1x4": "bhm_spiders_1x4",
    "dark_flexible_2x2": "bhm_spiders_1x4",
    "bright_3x1": "bhm_boss_bright_3x1",
    "bright_flexible_2x1": "bhm_boss_bright_2x1",
}


class BhmAqmesBaseCarton(BaseCarton):
    """
    Parent class that provides the underlying selections for all AQMES cartons
    """

    name = "bhm_aqmes_base"
    category = "science"
    mapper = "BHM"
    program = "bhm_aqmes"
    instrument = "BOSS"
    inertial = True
    tile = False
    priority = None
    alias_c = None
    alias_t = None
    alias_c2s = None
    cadence = None
    cadence_v0p5 = None

    def get_fieldlist(self, cadence_v0=None):
        """
        read the AQMES field centres from a fits file and convert to a list of dicts
        New: it is now the responsibility of the calling function to convert from a
        v1 cadence into a v0 cadence name
        """
        stub = self.parameters.get("fieldlist", None)
        if stub is None or stub == "" or stub == "None":
            return None

        # filename = pkg_resources.resource_filename( __name__, stub)
        filename = str(importlib.resources.files("target_selection") / stub)
        assert len(filename) > 0

        try:
            hdul = fits.open(filename)
        except BaseException:
            raise Exception(f"Failed to find/open fieldlist file: {filename}")

        assert len(hdul[1].data) > 0

        # # choose the correct subset of fields based on the cadence name
        # # and then form a list of dicts
        # # we have to use the v0 cadence names in the file though
        # assert cadence_v1 in cadence_map_v1_to_v0
        # cadence_v0 = cadence_map_v1_to_v0[cadence_v1]

        try:
            fieldlist = [
                {
                    "racen": r["RACEN"],
                    "deccen": r["DECCEN"],
                    "radius": radius_apo,
                }
                for r in hdul[1].data
                if r["CADENCE"] == cadence_v0
            ]
        except BaseException:
            raise Exception(f"Error interpreting contents of fieldlist file: {filename}")

        assert len(fieldlist) > 0

        return fieldlist

    def append_spatial_query(self, query, fieldlist):
        """Extend the peewee query using a list of field centres"""
        if fieldlist is None:
            return query
        elif len(fieldlist) == 0:
            return query

        q = False
        for f in fieldlist:
            q = q | peewee.fn.q3c_radial_query(
                self.alias_c.ra, self.alias_c.dec, f["racen"], f["deccen"], f["radius"]
            )
        return query.where(q)

    # main query
    def build_query(self, version_id, query_region=None):
        c = Catalog.alias()
        c2s = CatalogFromSDSS_DR19p_Speclite.alias()
        s = SDSS_DR19p_Speclite.alias()
        t = SDSS_DR16_QSO.alias()
        self.alias_c = c
        self.alias_t = t
        self.alias_c2s = c2s

        # set the Carton priority+values here - read from yaml
        priority_floor = peewee.Value(int(self.parameters.get("priority", 999999)))
        value = peewee.Value(self.parameters.get("value", 1.0)).cast("float")
        instrument = peewee.Value(self.instrument)
        inertial = peewee.Value(self.inertial).cast("bool")
        opt_prov = peewee.Value("sdss_psfmag")
        # for v1 read cadence from param file rather than from class code
        cadence_v1 = self.parameters.get("cadence", "unknown")
        cadence = peewee.Value(cadence_v1).cast("text")
        if hasattr(self, "cadence_v0"):
            cadence_v0 = peewee.Value(self.cadence_v0)
        else:
            cadence_v0 = peewee.Value(cadence_map_v1_to_v0[cadence_v1]).cast("text")
        # cadence_v0 = peewee.Value(cadence_map_v0p5_to_v0[self.cadence_v0p5]).cast('text')
        # cadence = peewee.Value(self.cadence_v0p5).cast('text')

        priority = priority_floor

        magnitude_sdss_g = peewee.Case(
            None, ((t.psfmag[1].between(0.1, 29.9), t.psfmag[1]),), "NaN"
        ).cast("float")
        magnitude_sdss_r = peewee.Case(
            None, ((t.psfmag[2].between(0.1, 29.9), t.psfmag[2]),), "NaN"
        ).cast("float")
        magnitude_sdss_i = peewee.Case(
            None, ((t.psfmag[3].between(0.1, 29.9), t.psfmag[3]),), "NaN"
        ).cast("float")
        magnitude_sdss_z = peewee.Case(
            None, ((t.psfmag[4].between(0.1, 29.9), t.psfmag[4]),), "NaN"
        ).cast("float")
        magnitude_gaia_g = peewee.Case(
            None, ((t.gaia_g_mag.between(0.1, 29.9), t.gaia_g_mag),), "NaN"
        ).cast("float")
        magnitude_gaia_bp = peewee.Case(
            None, ((t.gaia_bp_mag.between(0.1, 29.9), t.gaia_bp_mag),), "NaN"
        ).cast("float")
        magnitude_gaia_rp = peewee.Case(
            None, ((t.gaia_rp_mag.between(0.1, 29.9), t.gaia_rp_mag),), "NaN"
        ).cast("float")

        query = (
            c.select(
                c.catalogid,
                t.pk.alias("dr16q_pk"),  # extra
                s.pk.alias("dr19p_pk"),  # extra
                c.ra,  # extra
                c.dec,  # extra
                t.ra.alias("dr16q_ra"),  # extra
                t.dec.alias("dr16q_dec"),  # extra
                priority.alias("priority"),
                value.alias("value"),
                inertial.alias("inertial"),
                instrument.alias("instrument"),
                cadence.alias("cadence"),
                cadence_v0.alias("cadence_v0"),  # extra
                opt_prov.alias("optical_prov"),
                magnitude_sdss_g.alias("g"),
                magnitude_sdss_r.alias("r"),
                magnitude_sdss_i.alias("i"),
                magnitude_sdss_z.alias("z"),
                magnitude_gaia_g.alias("gaia_g"),
                magnitude_gaia_bp.alias("bp"),
                magnitude_gaia_rp.alias("rp"),
                t.plate.alias("dr16q_plate"),  # extra
                t.mjd.alias("dr16q_mjd"),  # extra
                t.fiberid.alias("dr16q_fiberid"),  # extra
                t.gaia_ra.alias("dr16q_gaia_ra"),  # extra
                t.gaia_dec.alias("dr16q_gaia_dec"),  # extra
                t.sdss2gaia_sep.alias("dr16q_sdss2gaia_sep"),  # extra
                t.z.alias("dr16q_redshift"),  # extra
                c2s.best.alias("c2s_best"),  # extra
                s.pk.alias("s19p_pk"),  # extra
                s.specprimary.alias("s19p_specprimary"),  # extra
            )
            .join(c2s)
            .join(s)
            .join(t, on=((s.plate == t.plate) & (s.mjd == t.mjd) & (s.fiberid == t.fiberid)))
            .where(
                c.version_id == version_id,
                c2s.version_id == version_id,
                c2s.best >> True,  # TODO check this is working in v1.0
                #                   # - this condition killed many AQMES
                #                   #   targets in v0+v0.5 cross-matches
            )
            .where(
                t.psfmag[3] >= self.parameters["mag_i_min"],
                t.psfmag[3] < self.parameters["mag_i_max"],
            )
            # .distinct([t.pk])   # avoid duplicates - trust the QSO parent sample
            .distinct([c.catalogid])  # avoid duplicates - trust the catalog
        )

        # query = self.append_spatial_query(query, self.get_fieldlist(cadence_v1))
        query = self.append_spatial_query(query, self.get_fieldlist(cadence_v0))

        return query


# -------AQMES medium section ------ #


class BhmAqmesMedCarton(BhmAqmesBaseCarton):
    """
    SELECT * FROM sdss_dr16_qso
    WHERE  psfmag_i BETWEEN 16.x AND 19.1
    AND   {target lies in spatial selection}
    """

    name = "bhm_aqmes_med"
    # cadence_v0p5 = 'dark_10x4_4yr'

    # TD's note to self:
    # add something like the following if want to add carton-specific selections
    #    def build_query(self, version_id, query_region=None):
    #        query = super().build_query(version_id, query_region)
    #        query = query.where( # .... add extra terms here
    #        )
    #        return query


class BhmAqmesMedFaintCarton(BhmAqmesBaseCarton):
    """
    SELECT * FROM sdss_dr16_qso
    WHERE  psfmag_i BETWEEN 19.1 AND 21.0
    AND   {target lies in spatial selection}
    """

    name = "bhm_aqmes_med_faint"
    # cadence_v0p5 = 'dark_10x4_4yr'
    program = "bhm_filler"


# -------AQMES medium section ----- #
#
#

#
# -------AQMES wide section ------ #


# class BhmAqmesWide3Carton(BhmAqmesBaseCarton):
#     '''
#     SELECT * FROM sdss_dr16_qso WHERE psfmag_i BETWEEN 16.0 AND 19.1
#     '''
#     name = 'bhm_aqmes_wide3'
#     cadence_v0p5 = 'dark_3x4'
#
#
# class BhmAqmesWide3FaintCarton(BhmAqmesBaseCarton):
#     '''
#     SELECT * FROM sdss_dr16_qso WHERE psfmag_i BETWEEN 19.1 AND 21.0
#     '''
#     name = 'bhm_aqmes_wide3_faint'
#     cadence_v0p5 = 'dark_3x4'
#     program = 'bhm_filler'


class BhmAqmesWide2Carton(BhmAqmesBaseCarton):
    """
    SELECT * FROM sdss_dr16_qso WHERE psfmag_i BETWEEN 16.0 AND 19.1
    """

    name = "bhm_aqmes_wide2"
    # cadence_v0p5 = 'dark_2x4'


class BhmAqmesWide2FaintCarton(BhmAqmesBaseCarton):
    """
    SELECT * FROM sdss_dr16_qso WHERE psfmag_i BETWEEN 19.1 AND 21.0
    """

    name = "bhm_aqmes_wide2_faint"
    # cadence_v0p5 = 'dark_2x4'
    program = "bhm_filler"


class BhmAqmesWide1Carton(BhmAqmesBaseCarton):
    """
    SELECT * FROM sdss_dr16_qso WHERE psfmag_i BETWEEN 16.0 AND 19.1
    """

    name = "bhm_aqmes_wide1"
    cadence_v0 = "bhm_aqmes_wide_2x4"


class BhmAqmesWide1FaintCarton(BhmAqmesBaseCarton):
    """
    SELECT * FROM sdss_dr16_qso WHERE psfmag_i BETWEEN 19.1 AND 21.0
    """

    name = "bhm_aqmes_wide1_faint"
    cadence_v0 = "bhm_aqmes_wide_2x4"
    program = "bhm_filler"


# -------AQMES wide section ------ #


# -------AQMES bonus section ------ #


class BhmAqmesBonusCoreCarton(BhmAqmesBaseCarton):
    """
    SELECT * FROM sdss_dr16_qso WHERE psfmag_i BETWEEN 16.0 AND 19.1
    {NO spatial constraint}
    """

    name = "bhm_aqmes_bonus_core"
    # cadence_v0p5 = 'dark_1x4'
    program = "bhm_filler"


class BhmAqmesBonusFaintCarton(BhmAqmesBaseCarton):
    """
    SELECT * FROM sdss_dr16_qso WHERE psfmag_i BETWEEN 19.1 AND 21.0
    """

    name = "bhm_aqmes_bonus_faint"
    # cadence_v0p5 = 'dark_1x4'
    program = "bhm_filler"


class BhmAqmesBonusBrightCarton(BhmAqmesBaseCarton):
    """
    SELECT * FROM sdss_dr16_qso WHERE psfmag_i BETWEEN 14.0 AND 18.0
    """

    name = "bhm_aqmes_bonus_bright"
    # cadence_v0p5 = 'bright_3x1'
    program = "bhm_filler"


# ------- AQMES bonus section ------ #
